import { Map } from "immutable"

export type ErrorMessage = string;

export type ErrorsMap = Map<string, ErrorMessage>


/**
 * How long should an error be displayed before dismissing
 */
export const ERROR_DISPLAY_TIME = 3000

/**
 * creates a new instance or an ErrorsMap
 */
export function newErrorsMap(): ErrorsMap {
    return Map<string, ErrorMessage>()
}

/**
 * Tries to add a new error message for the given ErrorsMap. 
 * If the error message is null, this function removes the error for this element from the errors map.
 * 
 * @param errors The original errors map
 * @param elementIdOfCausedError The id (like uml type id, or uml method id , etc. ) that has caused the error
 * @param msg The error message. if null the previous error message for this element will be removed
 */
export function addOrRemoveError(errors: ErrorsMap, elementIdOfCausedError: string, msg: ErrorMessage): ErrorsMap {
    const oldError = errors.get(elementIdOfCausedError)
    if (oldError == msg) {
        // Nothing has changed. 
        // Either:
        //  - same error as before or 
        //  - there was no error before and now there is no new one (both are null or undefined)
        return errors
    }

    if (msg != null) {
        // There is a new error
        return errors.set(elementIdOfCausedError, msg)
    }

    return errors.remove(elementIdOfCausedError)
}

/**
 * Remove errors of a certain type. Usually this must be invoked when the coresponding element (like uml class or uml method etc.) has been removed
 * @param errors the original errors map
 * @param elementIdOfCausedError The id (like uml class id, or uml method id etc.) for which you want to remove all errors
 */
export function removeError(errors: ErrorsMap, elementIdOfCausedError): ErrorsMap {
    return addOrRemoveError(errors, elementIdOfCausedError, null)
}