import { Map } from "immutable"

/**
 * Just a type alias
 */
export type SelectedElementsMap = Map<string, SelectedElement>

/**
 * Creates a new instance
 */
export function newSelectedElementsMap(): SelectedElementsMap {
    return Map<string, SelectedElement>()
}

/**
 * 
 * @param previous The previous selected elements. This is a memory optimization as it will return this one if
 * previous is empty
 */
export function noSelectedElements(previous: SelectedElementsMap): SelectedElementsMap {
    if (previous.isEmpty()) {
        return previous
    }
    else
        return newSelectedElementsMap()
}

/**
 * Only selects the given element
 * @param selectedElement The new selected element 
 */
export function setSelected(selectedElement: SelectedElement) {
    return newSelectedElementsMap().set(selectedElement.elementId, selectedElement);
}

/**
 * Only selects the given list of elements
 * @param selectedElements list (varargs) of Selected elements
 */
export function setMultipleSelected(...selectedElements: SelectedElement[]) {
    return newSelectedElementsMap().withMutations(map => {
        selectedElements.forEach(selected => map.set(selected.elementId, selected))
    });
}

/**
 * Checks if a property is selected and editing
 * @param map The selected elements map
 * @param umlPropertyId the id of the EditorUmlProperty
 * @return a pair. First element is for "selected", second is for "is editing".
 */
export function isProperySelected(map: SelectedElementsMap, umlPropertyId: string): [boolean, boolean] {
    const found = map.get(umlPropertyId)
    if (found == null) {
        return [false, false]
    }
    return [true, (found as PropertySelectedElement).editing]

}

/**
 * Checks if a method is selected and editing
 * @param map The selected elements map
 * @param umlMethodId the id of the EditorUmlMethod
 * @return a pair. First element is for "selected", second is for "is editing".
 */
export function isMethodSelected(map: SelectedElementsMap, umlMethodId: string): [boolean, boolean] {
    const found = map.get(umlMethodId)
    if (found == null) {
        return [false, false]
    }
    return [true, (found as MethodSelectedElement).editing]
}

/**
 * Get the PropertySelectedElement or MethodSelectedElement. Returns null if no property or method is selected
 * @param map 
 */
export function getSelectedPropertyOrMethod(map: SelectedElementsMap): MethodSelectedElement | PropertySelectedElement {
    if (map.size == 2) {
        // If a property or method is selected map contains exactly 2 selected items: 
        // - The method or property itself
        // - The surrounding parent
        const found = map.filter(i => i instanceof PropertySelectedElement || i instanceof MethodSelectedElement)
            .first()

        if (found instanceof MethodSelectedElement)
            return found
        else if (found instanceof PropertySelectedElement)
            return found
    }

    return null
}


/**
 * Checks if a UmlType (like class or interface) is selected and editing the title
 * @param map The selected elements map
 * @param umlTypeId the id of the EditorUmlType
 * @return a pair. First element is for "selected", second is for "is editing the name".
 */
export function isTypeSelected(map: SelectedElementsMap, umlTypeId: string): [boolean, boolean] {
    const found = map.get(umlTypeId)
    if (found == null) {
        return [false, false]
    }
    return [true, (found as UmlTypeSelected).editingName]
}


/**
 * It's a class to be able to use instance of checks later
 */
export class SelectedElement {

    /** 
     * The id of the element (like UmlClass, UmlMethod etc.) that has been selected
     */
    readonly elementId: string

    constructor(elementId: string) {
        this.elementId = elementId
    }
}

/**
 * A UmlType is selected. Also it holds the info whether or not the uml type name is currently in editing mode
 */
export class UmlTypeSelected extends SelectedElement {
    /**
     * Is the user currently editing the name of the UmlType (like class or interface).
     * So this means, not only is the UmlType selected, but also the UmlType name is in edit mode
     * which means the user want to edit the name
     */
    readonly editingName: boolean
    constructor(elementId: string, editingName: boolean) {
        super(elementId)
        this.editingName = editingName
    }
}

/**
 * Indicates that a Method of a umltype has been selected
 */
export class MethodSelectedElement extends SelectedElement {
    /**
     * The id of the "parent" UmlType this method belongs to
     */
    readonly umlTypeId: string

    /**
     * Is the user currently editing the method specification
     */
    readonly editing: boolean

    constructor(elementId: string, umlTypeId: string, editing: boolean) {
        super(elementId)
        this.umlTypeId = umlTypeId
        this.editing = editing;
    }
}

/**
 * Indicates that a Property of a UmlClass has been selected
 */
export class PropertySelectedElement extends SelectedElement {
    /**
    * The id of the "parent" UmlType this method belongs to
    */
    readonly umlTypeId: string
    /**
     * Is the user currently editing the property specification
     */
    readonly editing: boolean

    constructor(elementId: string, umlTypeId: string, editing: boolean) {
        super(elementId)
        this.umlTypeId = umlTypeId
        this.editing = editing
    }
}

/**
 * Indicates that a UmlRelation is selected and whether or not the user is editing multiplicity
 */
export class SelectUmlRelation extends SelectedElement {
    readonly editingStartMultiplicity : boolean
    readonly editingEndMultiplicity : boolean
    constructor(elementId: string, editingStartMultiplicity : boolean, editineEndMultiplicity: boolean){
        super(elementId)
        this.editingEndMultiplicity = editineEndMultiplicity
        this.editingStartMultiplicity = editingStartMultiplicity
    }
}