import { Intent } from "./Intent"
import { InteractionTransaction } from "../interaction/InteractionTransaction"
import { EditorUmlType } from "../model/EditorUmlClass"
import { EditorUmlClassProperty } from "../model/EditorUmlClassProperty"
import { EditorUmlTypeMethod } from "../model/EditorUmlTypeMethod"
import { UmlConnectionType } from "../../components/uml/UmlConnectionCanvas"

/**
 * Means a type that can be selected 
 */
export type SelectableElement = EditorUmlType

/**
 * Indicates that somewhere on the UML Canvas has been clicked. This typically means, cancel all interactions and 
 * deselect all elements
 */
export class ClickOnUmlCanvasIntent implements Intent {
    readonly currentInteraction: InteractionTransaction;
    readonly x: number;
    readonly y: number;

    constructor(currentInteraction: InteractionTransaction, x: number, y: number) {
        this.currentInteraction = currentInteraction;
        this.x = x;
        this.y = y;
    }

    toString(): string {
        return "ClickOnUmlCanvasIntent at " + this.x + " / " + this.y + " . Current Interaction " + this.currentInteraction;
    }
}

/**
 * Intent to select a single UmlType
 */
export class SelectUmlTypeIntent implements Intent {
    readonly umlType: EditorUmlType

    constructor(umlType: EditorUmlType) {
        this.umlType = umlType
    }

    toString(): string {
        return "SelecteUmlTypeIntent: Element selected: " + this.umlType.name;
    }
}

/**
 * Intent to select a single UmlType and the given uml class property
 */
export class SelectUmlTypeAndPropertyIntent implements Intent {
    readonly umlType: EditorUmlType
    readonly umlPropertyId: string

    constructor(umlType: EditorUmlType, umlPropertyId: string) {
        this.umlType = umlType
        this.umlPropertyId = umlPropertyId
    }

    toString(): string {
        return "SelecteUmlTypeAndPropertyIntent: Element selected: " + this.umlType.name + " property: " + this.umlPropertyId;
    }
}



/**
 * Intent to select a single UmlType and the given uml method
 */
export class SelectUmlTypeAndMethodIntent implements Intent {
    readonly umlType: EditorUmlType
    readonly umlMethodId: string

    constructor(umlType: EditorUmlType, umlMethodId: string) {
        this.umlType = umlType
        this.umlMethodId = umlMethodId
    }

    toString(): string {
        return "SelecteUmlTypeAndMethodIntent: Element selected: " + this.umlType.name + " method: " + this.umlMethodId
    }
}


/**
 * Intent to select a single UmlType and the given uml class property
 */
export class SelectUmlInheritanceIntent implements Intent {
    readonly umlInheritanceId: string

    constructor(umlInheritanceId: string) {
        this.umlInheritanceId = umlInheritanceId
    }

    toString(): string {
        return "SelectUmlInheritanceIntent: Element selected: " + this.umlInheritanceId;
    }
}

/**
 * Intent to select a single UmlType and the given uml class property
 */
export class SelectUmlInterfaceImplementationIntent implements Intent {
    readonly umlInterfaceImplementation: string

    constructor(umlInterfaceImplementation: string) {
        this.umlInterfaceImplementation = umlInterfaceImplementation
    }

    toString(): string {
        return "SelectUmlInterfaceImplementationIntent: Element selected: " + this.umlInterfaceImplementation;
    }
}


/**
 * Intent to select a single UmlRelation
 */
export class SelectUmlRelationIntent implements Intent {
    readonly umlRelationId: string

    constructor(umlRelationId: string) {
        this.umlRelationId = umlRelationId
    }

    toString(): string {
        return "SelectUmlRelationIntent: Element selected: " + this.umlRelationId;
    }
}


export enum UmlRelationTextFieldType {
    START_MULTIPLICITY,
    END_MULTIPLICITY,
    START_NAME,
    END_NAME
}

/**
 * Intent to select a single UmlRelation and also start editing the relation
 */
export class StartEditingUmlRelationTextFieldIntent implements Intent {
    readonly umlRelationId: string
    readonly relationField: UmlRelationTextFieldType

    constructor(umlRelationId: string, relationField: UmlRelationTextFieldType) {
        this.umlRelationId = umlRelationId
        this.relationField = relationField
    }

    toString(): string {
        return "StartEditingUmlRelationTextFieldIntent: Element selected: " + this.umlRelationId + " " + this.relationField;
    }
}

/**
 * The intent to cancel editing a uml relation multiplicity
 */
export class CancelEditingUmlRelationTextFieldIntent implements Intent {
    readonly umlRelationId: string
    readonly relationField: UmlRelationTextFieldType

    constructor(umlRelationId: string, relationField: UmlRelationTextFieldType) {
        this.umlRelationId = umlRelationId
        this.relationField = relationField
    }

    toString(): string {
        return "CancelEditingUmlRelationTextFieldIntent: Element: " + this.umlRelationId + " " + this.relationField;
    }
}

/**
 * The intent to cancel editing a uml relation multiplicity
 */
export class ChangeUmlRelationTextFieldIntent implements Intent {
    readonly umlRelationId: string
    readonly relationField: UmlRelationTextFieldType
    readonly updatedValue: string

    constructor(umlRelationId: string, relationField: UmlRelationTextFieldType, updatedValue: string) {
        this.umlRelationId = umlRelationId
        this.relationField = relationField
        this.updatedValue = updatedValue
    }

    toString(): string {
        return "CancelEditingUmlRelationTextFieldIntent: Element: " + this.umlRelationId + " " + this.relationField + " updated value: " + this.updatedValue;
    }
}




/**
 * Intent to doubleclick an UmlConnection
 */
export class DoubleClickUmlConnectionIntent implements Intent {
    readonly umlConnectionId: string
    readonly connectionType: UmlConnectionType
    readonly position: { x: number, y: number }

    constructor(umlConnectionId: string, x: number, y: number, connectionType: UmlConnectionType) {
        this.umlConnectionId = umlConnectionId
        this.position = { x: x, y: y }
        this.connectionType = connectionType
    }

    toString(): string {
        return "DoubleClickUmlConnectionIntent: Element doubleclicked: " + this.umlConnectionId;
    }
}


/**
 * Intent to delete the selected elements (can also be just one single element)
 */
export class DeleteSelectedElementsIntent implements Intent {

    toString(): string {
        return "DeleteSelectedElementsIntent";
    }
}

/**
 * The intent to move the currently selected elements on the canvas.
 * For lower memory usage internally this class uses a pool.
 * So you have to use static method obtain() to get a instance and call recycle() once you are done 
 * consuming this intent
 */
export class MoveSelectedElementsIntent implements Intent {


    static POOL: MoveSelectedElementsIntent[] = []


    newX: number
    newY: number

    private constructor(newX: number, newY: number) {
        this.newX = newX;
        this.newY = newY
    }

    recycle() {
        this.newX = 0
        this.newY = 0
        MoveSelectedElementsIntent.POOL.push(this)
    }

    static obtain(newX: number, newY: number): MoveSelectedElementsIntent {
        if (MoveSelectedElementsIntent.POOL.length == 0) {
            return new MoveSelectedElementsIntent(newX, newY);
        } else {
            const reused = MoveSelectedElementsIntent.POOL.pop()
            reused.newX = newX
            reused.newY = newY
            return reused
        }
    }


    toString(): string {
        return "MoveSelectedElementsIntent. newX=" + this.newX + " newY=" + this.newY;
    }

}




/**
 * The intent to move the currently selected elements on the canvas.
 * For lower memory usage internally this class uses a pool.
 * So you have to use static method obtain() to get a instance and call recycle() once you are done 
 * consuming this intent
 */
export class MoveAnchorPointIntent implements Intent {


    static POOL: MoveAnchorPointIntent[] = []

    umlConnectionId: string
    newPercentage: number
    /**
     * True means the fist AnchorPoint of the uml connection is moved, false means the second one
     */
    firstAnchorPoint: boolean

    private constructor(umlConnectionId: string, newPercentage: number, firstAnchorPoint: boolean) {
        this.umlConnectionId = umlConnectionId
        this.newPercentage = newPercentage
        this.firstAnchorPoint = firstAnchorPoint
    }

    recycle() {
        MoveAnchorPointIntent.POOL.push(this)
    }

    static obtain(umlConnectionId: string, newPercentage: number, firstAnchorPoint: boolean): MoveAnchorPointIntent {
        if (MoveAnchorPointIntent.POOL.length == 0) {
            return new MoveAnchorPointIntent(umlConnectionId, newPercentage, firstAnchorPoint);
        } else {
            const reused = MoveAnchorPointIntent.POOL.pop()
            reused.umlConnectionId = umlConnectionId
            reused.newPercentage = newPercentage
            reused.firstAnchorPoint = firstAnchorPoint
            return reused
        }
    }


    toString(): string {
        return "MoveAnchorPointIntent. connection id : " + this.umlConnectionId + " " + this.newPercentage + "%" + " " + this.firstAnchorPoint;
    }
}


/**
 * Intent to start moving an AnchorPoint of a UML connection like inheritance, association etc.
 */
export class StartMovingAnchorPointIntent implements Intent {
    readonly firstAnchorPoint: boolean
    readonly umlConnectionId: string
    readonly minMouseValue: number
    readonly maxMouseValue: number

    /**
     * true means moving is only allowed on the X achsis, false means allowed on Y achsis
     */
    readonly achsis: boolean

    constructor(umlConnectionId: string, firstAnchorPoint: boolean, minMouseValue: number, maxMouseValue: number, achsis: boolean) {
        this.firstAnchorPoint = firstAnchorPoint
        this.umlConnectionId = umlConnectionId
        this.minMouseValue = minMouseValue
        this.maxMouseValue = maxMouseValue
        this.achsis = achsis
    }


    toString(): string {
        return "StartMovingAnchorPointIntent. connection id : " + this.umlConnectionId + " " + this.firstAnchorPoint;
    }
}


/**
 * Intent to stop any ongoing moving AnchorPoint action
 */
export class StopMovingAnchorPointIntent implements Intent {

    toString(): string {
        return "StopMovingAnchorPointIntent";
    }
}


/**
 * Intent to cancel all ongoing stuff. Usually this is invoked by pressing the ESC key.
 */
export class CancelEverythingIntent implements Intent {

    toString(): string {
        return "CancelEverythingIntent";
    }
}