import { UmlType } from "../model/EditorUmlClass"
import { MenuItemType } from "../MenuItem"

/**
 * Allows to build user interaction where multiple actions are required one after another.
 * For example to create a UML Type, first you have to press the corresponding menu button which will create a new transaction.
 * Then you have to click somewhere on the canvas to determine the position where the new UML TYPE should be placed.
 * Taking track of this two "click" events is the responsibility of a InteractionTransaction
 */
export interface InteractionTransaction {
}

/**
 * Returns true if the passed InteractionTransaction represents  "no InteractionTransaction runnining right now". 
 * Otherwise, false.
 * @param i the transcation that should be checked
 */
export function isNoInteractionRunning(i: InteractionTransaction) {
    return i == NoInteractionTransactionRunning.getInstance();
}

/**
 * Indicates that no interaction is currently running.
 * This is a singleton. Use getInstance() method.
 */
export class NoInteractionTransactionRunning implements InteractionTransaction {

    private static readonly _instance = new NoInteractionTransactionRunning();

    private constructor() { }

    static getInstance(): NoInteractionTransactionRunning {
        return NoInteractionTransactionRunning._instance;
    }

    toString(): string {
        return "NoInteractionTransactionRunning";
    }
}

/**
 * Indidcates that currently there is no special form of interaction running
 */
export class AddTypeInteractionTransaction implements InteractionTransaction {
    readonly umlType: UmlType;
    constructor(umlType: UmlType) {
        this.umlType = umlType;
    }

    toString(): string {
        return "AddTypeInteractionTransaction " + this.umlType;
    }
}


/**
 * Indicates that the user should select the super class to extend from.
 * The informtion about the previously selected class will also be hold in here
 */
export class SelectingTwoElementsInteractionTransaction implements InteractionTransaction {

    readonly menuItemType: MenuItemType;
    /**
     * Is null if no element has been selected yet
     */
    readonly firstSelectedElementId: string;

    constructor(menuItemType: MenuItemType, firstSelectedElementId: string) {
        this.firstSelectedElementId = firstSelectedElementId;
        this.menuItemType = menuItemType;
    }

    toString(): string {
        return "SelectingTwoElementsInteractionTransaction ";
    }
}


/**
 * InteractionTransaction that indicates that a AnchorPoint is currently moved
 */
export class MovingAnchorPointInteractionTransaction implements InteractionTransaction {
    readonly firstAnchorPoint: boolean
    readonly umlConnectionId: string
    readonly minMouseValue: number
    readonly maxMouseValue: number

    /**
     * true means moving is only allowed on the X achsis, false means allowed on Y achsis
     */
    readonly achsis: boolean

    constructor(umlConnectionId: string, firstAnchorPoint: boolean, minMouseValue: number, maxMouseValue: number, achsis: boolean) {
        this.firstAnchorPoint = firstAnchorPoint
        this.umlConnectionId = umlConnectionId
        this.minMouseValue = minMouseValue
        this.maxMouseValue = maxMouseValue
        this.achsis = achsis
    }


    toString(): string {
        return "MovingAnchorPointInteractionTransaction. UmlConnectionId " + this.umlConnectionId + " " + this.firstAnchorPoint;
    }
}