import { EditorUmlDocument } from './EditorUmlDocument'
import { EditorUmlType, UmlType } from './EditorUmlClass'
import { EditorUmlClassProperty } from './EditorUmlClassProperty'
import { EditorUmlInheritance } from './EditorUmlInheritance'
import { EditorUmlInterfaceImplementation } from './EditorUmlInterfaceImplementation'
import { EditorUmlTypeMethod } from './EditorUmlTypeMethod'
import { EditorUmlRelation, relationTypeFromString } from './EditorUmlRelation'
import { AnchorPoint } from './AnchorPoint'
import { Map, OrderedMap, OrderedSet } from "immutable"

const VISIBILITY_PUBLIC = "PUBLIC";
const VISIBILITY_PRIVATE = "PRIVATE";


/**
 * Converts a json object to EditorUmlDocument
 * @param json the json object representation of the whole uml document
 */
export function umlDocumentFromJson(json: any): EditorUmlDocument {

    const classes = OrderedMap<string, EditorUmlType>().withMutations(map => {
        json.classes.forEach(c => {
            const cls = umlTypeFromJson(c)
            map.set(cls.id, cls)
        })

        json.interfaces.forEach(c => {
            const inf = umlTypeFromJson(c)
            map.set(inf.id, inf)
        })
    })

    const inheritance = Map<string, EditorUmlInheritance>().withMutations(map => {
        json.classInheritance.forEach(element => {
            const inh = umlInheritanceFromJson(element)
            map.set(inh.id, inh)
        });

        json.interfaceInheritance.forEach(element => {
            const inh = umlInheritanceFromJson(element)
            map.set(inh.id, inh)
        });
    })

    const classImplementingInterface = Map<string, EditorUmlInterfaceImplementation>().withMutations(map => {
        json.classImplementingInterface.forEach(element => {
            const impl = umlInterfaceImplementationFormJson(element)
            map.set(impl.id, impl)
        });
    })

    const relations = Map<string, EditorUmlRelation>().withMutations(map => {
        json.relations.forEach(element => {
            const rel = umlRelationFromJson(element)
            map.set(rel.id, rel)
        });
    })

    return {
        types: classes,
        inheritance: inheritance,
        classImplementingInterface: classImplementingInterface,
        relations: relations
    };
}

/**
 * Converts a json object to an EditorUmlType
 * @param json the json object representation of the uml type (either interface or class)
 */
function umlTypeFromJson(json: any): EditorUmlType {

    const type = json.properties === undefined ? UmlType.INTERFACE : UmlType.CLASS;
    const methods = OrderedMap<string, EditorUmlTypeMethod>().withMutations(map => {
        json.methods.forEach(m => {
            const method = umlMethodFromJson(m);
            map.set(method.id, method);
        });
    });


    let props = OrderedMap<string, EditorUmlClassProperty>();
    if (type === UmlType.CLASS) {
        props = props.withMutations(map => {
            json.properties.forEach(p => {
                const property = umlPropertyFromJson(p);
                map.set(property.id, property);
            });
        });
    }


    return {
        id: json.meta.id,
        name: json.name,
        type: type,
        methods: methods,
        properties: props,
        x: json.meta.x,
        y: json.meta.y
    };
}

/**
 * Converts the string value to a uml visibilityl notation "+" for public "-" for private
 * 
 * @param visibility Converts the strign value to a uml notation representation
 */
function getVisibilityModifier(visibility: string): string {
    return visibility === VISIBILITY_PUBLIC ? "+" : "-"
}

/**
 * Converts a json object to a EditorUmlTypeMethod
 * @param json to original json object representation of a uml method
 */
function umlMethodFromJson(json: any): EditorUmlTypeMethod {

    let visibility = getVisibilityModifier(json.visibility)
    let params = "";
    for (let i = 0; i < json.parameters.length; i++) {
        let par = json.parameters[i];
        if (i > 0) {
            params = params + ", "
        }
        params = params + par.name + " : " + par.type;
    }

    let returnType = ""
    if (json.returnType.length > 0) {
        returnType = " : " + json.returnType;
    }

    let methodSpec = visibility + " " + json.name + "(" + params + ")" + returnType;

    return {
        id: json.meta.id,
        methodSepcification: methodSpec
    };
}


/**
 * Converts a json object into a EditorUmlClassProperty
 * @param json the original json representation of a uml propertySpec
 */
function umlPropertyFromJson(json: any): EditorUmlClassProperty {
    return {
        id: json.meta.id,
        propertySpec: getVisibilityModifier(json.visibility) + " " + json.name + " : " + json.type
    }
}

function umlInheritanceFromJson(json: any): EditorUmlInheritance {
    return {
        id: json.meta.id,
        subTypeId: json.meta.subTypeId,
        superTypeId: json.meta.superTypeId,
        subTypeAnchorPoint: json.meta.subTypeAnchorPoint,
        superTypeAnchorPoint: json.meta.superTypeAnchorPoint,
        points: OrderedSet.of(json.meta.points)
    }
}


function umlInterfaceImplementationFormJson(json: any): EditorUmlInterfaceImplementation {
    return {
        id: json.meta.id,
        interfaceId: json.meta.interfaceId,
        classId: json.meta.classId,
        interfaceAnchorPoint: json.meta.interfaceAnchorPoint,
        classAnchorPoint: json.meta.classAnchorPoint,
        points: OrderedSet.of(json.meta.points)
    }
}

function umlRelationFromJson(json: any): EditorUmlRelation {

    return {
        id: json.meta.id,
        startTypeId: json.meta.startTypeId,
        endTypeId: json.meta.endTypeId,
        startMultiplicity: json.startMultiplicity,
        endMultiplicity: json.endMultiplicity,
        endTypeAnchorPoint: json.meta.endTypeAnchorPoint,
        startTypeAnchorPoint: json.meta.startTypeAnchorPoint,
        relationType: relationTypeFromString(json.relationType),
        points: OrderedSet.of(json.meta.points)
    }
}