import { Record, Map, OrderedMap } from "immutable"
import { EditorUmlType, UmlType } from './EditorUmlClass'
import { AnchorPoint, AnchorPointSide, calcAnchorPoints } from './AnchorPoint'
import { EditorUmlTypeMethod } from './EditorUmlTypeMethod'
import { EditorUmlClassProperty } from './EditorUmlClassProperty'
import { EditorUmlInterfaceImplementation } from './EditorUmlInterfaceImplementation'
import { EditorUmlInheritance } from './EditorUmlInheritance'
import { EditorUmlRelation, UmlRelationType } from './EditorUmlRelation'
import { SelectableElement } from '../intent/UmlCanvasIntents'
import { SelectedElementsMap } from '../SelectedElement'
import { UUID } from "angular2-uuid"
import { Set, OrderedSet } from "immutable"
import { ErrorMessage } from "../ErrorMessage"
import { Point } from "./Point"

/**
 * Represents the a UML Document, holding UML Classes, interfaces and other relations
 */
export type EditorUmlDocument = {
    readonly types: Map<string, EditorUmlType>,
    readonly inheritance: Map<string, EditorUmlInheritance>
    readonly classImplementingInterface: Map<string, EditorUmlInterfaceImplementation>
    readonly relations: Map<string, EditorUmlRelation>
}


/**
 * Creates an empty UML document
 */
export function emptyUmlDocument(): EditorUmlDocument {
    return {
        types: Map<string, EditorUmlType>(),
        inheritance: Map<string, EditorUmlInheritance>(),
        classImplementingInterface: Map<string, EditorUmlInterfaceImplementation>(),
        relations: Map<string, EditorUmlRelation>()
    }
}


/**
 * Adds a EditorUmlClass to the uml document.
 * 
 * @param umlDocument the original umlDocument where we want to add the new className+
 * @param className The uml class name
 * @param x the x position where the uml class is located 
 * @param y the y position where the uml class is located
 */
export function addUmlType(umlDocument: EditorUmlDocument, id: string, className: string, x: number, y: number, umlType: UmlType): [EditorUmlDocument, ErrorMessage] {

    const types = umlDocument.types;
    const errorMessage = isValidUmlTypeName(umlDocument, className)
    const newClass: EditorUmlType = {
        id: id,
        name: className,
        type: umlType,
        methods: OrderedMap<string, EditorUmlTypeMethod>(),
        properties: OrderedMap<string, EditorUmlClassProperty>(),
        x: x,
        y: y
    };

    return [
        {
            types: types.set(newClass.id, newClass),
            inheritance: umlDocument.inheritance,
            classImplementingInterface: umlDocument.classImplementingInterface,
            relations: umlDocument.relations
        },
        errorMessage
    ];
}


/**
 * Checks if the name is a valid one
 * @param umlDocument the uml document to check
 * @param name The name of the class
 */
function isValidUmlTypeName(umlDocument: EditorUmlDocument, name: string): string {

    const nameRegEx = /^\w+$/g
    if (!nameRegEx.test(name)) {
        return "Name is not valid. Only alphanumeric characters and underscore is allowed";
    }

    const sameName = umlDocument.types.filter(c => c.name === name);
    if (!sameName.isEmpty())
        return "A type with the name '" + name + "' already exists. Name must be unique.";

    return null;
}

/**
 * Checks if a multiplicity definition of a uml relation is a valid string
 * @param multiplictyStr The multiplicity string
 */
export function isValidMultiplicity(multiplictyStr: string): string {
    const regEx = /^([1-9]+|\*|[0-9]+..([1-9]+|\*))$/g
    if (!regEx.test(multiplictyStr)) {
        return "Wrong multiplicity. Examples: 1 or * or 2..4 or 1..*"
    }

    const indexOfDot = multiplictyStr.indexOf("..")
    if (indexOfDot >= 0) {
        const lowerBound = Number(multiplictyStr.substring(0, indexOfDot))
        const upperBoundStr = multiplictyStr.substring(indexOfDot + 2)
        if (upperBoundStr !== "*") {
            if (lowerBound > Number(upperBoundStr)) {
                return "Lower bound (" + lowerBound + ") must be less than upper bound (" + upperBoundStr + ") in multiplicity defintion '" + multiplictyStr + "'"
            }
        }

        return null // valid multiplicity string
    }
}


/**
 * 
 * @param relationId The UML relation id
 * @param startMultiplicity true if we want to compute the id of the start multiplicity. false if we want to compute the end multiplicity.
 */
export function umlRelationMultiplicityId(relationId: string, startMultiplicity: boolean) {
    return relationId + "-" + (startMultiplicity === true ? "start-multiplicity" : "end-multiplicity")
}

/**
 * Rename a UML class
 * @param umlDocument the original uml document where we want to rename the uml class
 * @param umlClassId The id of the umlClassId
 * @param newName The new name of the umlclass
 */
export function renameUmlType(umlDocument: EditorUmlDocument, umlTypeId: string, newName: string): [EditorUmlDocument, ErrorMessage] {

    const umlType: EditorUmlType = umlDocument.types.get(umlTypeId);

    if (umlType !== undefined) {
        // Rename UML class
        const errorMessage = isValidUmlTypeName(umlDocument, newName)
        const renamedClass: EditorUmlType = {
            id: umlType.id,
            name: newName,
            type: umlType.type,
            methods: umlType.methods,
            properties: umlType.properties,
            x: umlType.x,
            y: umlType.y,
        };

        return [
            {
                types: umlDocument.types.set(umlType.id, renamedClass),
                inheritance: umlDocument.inheritance,
                classImplementingInterface: umlDocument.classImplementingInterface,
                relations: umlDocument.relations
            },
            errorMessage]
    } else {
        throw new Error("Couldn't find uml type with id = " + umlTypeId);
    }
}



/**
 * Add a uml method
 * @param umlDocument the original uml document that should be changed
 * @param umlTypeId the uuid of the uml type where we want to add a new uml method
 */
export function addUmlMethod(umlDocument: EditorUmlDocument, umlTypeId: string, methodId: string): [EditorUmlDocument, ErrorMessage] {

    const newMethod: EditorUmlTypeMethod = {
        id: methodId,
        methodSepcification: "+ method()",
    }

    const typeToChange = umlDocument.types.get(umlTypeId);
    if (typeToChange === undefined) {
        throw new Error("Couldn't find uml type with id " + umlTypeId);
    }
    const errorMessage = isValidMethodSpecification(typeToChange, newMethod.methodSepcification)
    const updatedMethods = typeToChange.methods.set(newMethod.id, newMethod);

    const changedType: EditorUmlType = {
        id: typeToChange.id,
        name: typeToChange.name,
        methods: updatedMethods,
        properties: typeToChange.properties,
        type: typeToChange.type,
        x: typeToChange.x,
        y: typeToChange.y
    }


    return [
        {
            types: umlDocument.types.set(changedType.id, changedType),
            inheritance: umlDocument.inheritance,
            classImplementingInterface: umlDocument.classImplementingInterface,
            relations: umlDocument.relations
        },
        errorMessage
    ]
}





/**
 * change the specification of a uml method
 * @param umlDocument The uml document
 * @param umlTypeId the uml type id
 * @param methodId the method id
 */
export function changeUmlMethodSpecification(umlDocument: EditorUmlDocument, umlTypeId: string, methodId: string, newMethodSpecification: string): [EditorUmlDocument, ErrorMessage] {

    const classToChange = umlDocument.types.get(umlTypeId);
    if (classToChange === undefined) {
        throw new Error("Couldn't find uml type with id " + umlTypeId);
    }


    const methodToChange: EditorUmlTypeMethod = classToChange.methods.get(methodId);

    if (methodToChange === undefined) {
        throw new Error("Couldn't find uml method with id " + methodId);
    }


    let errorMessage = isValidMethodSpecification(classToChange, newMethodSpecification);


    const updatedMethods = classToChange.methods.set(methodToChange.id, {
        id: methodToChange.id,
        methodSepcification: newMethodSpecification,
    });

    const changedClass: EditorUmlType = {
        id: classToChange.id,
        name: classToChange.name,
        type: classToChange.type,
        methods: updatedMethods,
        properties: classToChange.properties,
        x: classToChange.x,
        y: classToChange.y
    };

    return [
        {
            types: umlDocument.types.set(changedClass.id, changedClass),
            inheritance: umlDocument.inheritance,
            classImplementingInterface: umlDocument.classImplementingInterface,
            relations: umlDocument.relations
        },
        errorMessage
    ]
}

/**
 * Checks if a given string is a valid uml method definition
 */
function isValidMethodSpecification(umlType: EditorUmlType, methodSpec: string): string {
    const methodRegEx = /^ *[+|-]? *\w\w* *\(( *(\w{1,} *: *\w{1,})* *){0,1}( *, *(\w{1,} *: *\w{1,}) *)*\) *(: *\w{1,}){0,1} *$/g;
    if (!methodRegEx.test(methodSpec)) {
        return "This is not a valid method definition. Example: foo(a : Bar, b : Other) : int "
    }

    // TODO better checks for similarity of properties? maybe format imput beforehand (remove multiple whitespaces)
    const sameMethodsCount = umlType.methods
        .filter(m => m.methodSepcification == methodSpec)
        .size

    if (sameMethodsCount > 0) {
        return "The same method is alredy defined"
    }

    return null
}


/**
 * Checks if a property spec is valid. Returns null if valid, otherwise a string error message.
 * @param umlType The "parent" uml type of the property
 * @param propertySpec The property specification in UML syntax
 */
function isValidPropertyName(umlType: EditorUmlType, propertySpec: string): string {
    const regEx = /^ *[+|-]? *\w+ *: *\w+ *$/g;
    if (!regEx.test(propertySpec)) {
        return "Property is not valid. Valid example: - counter : int"
    }

    // TODO better checks for similarity of properties? maybe format imput beforehand (remove multiple whitespaces)
    const samePropertyCount = umlType.properties.filter(p => p.propertySpec == propertySpec).size
    if (samePropertyCount > 0) {
        return "Same property is aldrady defined."
    }
    return null;
}

/**
 * Add a new property to the given uml type.
 * 
 * @param umlDocument The original document
 * @param typeId The id of the uml type where we want to add a property
 * @param propertySpec the property specification
 */
export function addUmlProperty(umlDocument: EditorUmlDocument, typeId: string, propertyId: string, propertySpec: string): [EditorUmlDocument, ErrorMessage] {

    const classToChange = umlDocument.types.get(typeId);

    if (classToChange === undefined) {
        throw new Error("Could not find uml type with id " + typeId);
    }


    const errorMessage = isValidPropertyName(classToChange, propertySpec)

    const newProperty: EditorUmlClassProperty = {
        id: propertyId,
        propertySpec: propertySpec,
    }

    const changedClass: EditorUmlType = {
        id: classToChange.id,
        name: classToChange.name,
        x: classToChange.x,
        y: classToChange.y,
        methods: classToChange.methods,
        properties: classToChange.properties.set(newProperty.id, newProperty),
        type: classToChange.type
    }

    return [
        {
            types: umlDocument.types.set(changedClass.id, changedClass),
            inheritance: umlDocument.inheritance,
            classImplementingInterface: umlDocument.classImplementingInterface,
            relations: umlDocument.relations
        },
        errorMessage
    ]
}

/**
 * Change the uml property spec and exit "editing mode"
 * 
 * @param umlDocument The original document
 * @param typeId The id of the uml type where we want to add a property
 * @param propertyId the property id
 * @param newPropertySpec the new property spec in uml from like "- foo : int"
 */
export function changeUmlProperty(umlDocument: EditorUmlDocument, typeId: string, propertyId: string, newPropertySpec: string): [EditorUmlDocument, ErrorMessage] {

    const classToChange = umlDocument.types.get(typeId);

    if (classToChange === undefined) {
        throw new Error("Could not find uml type with id " + typeId);
    }

    const propertyToChange = classToChange.properties.get(propertyId)
    if (propertyToChange === undefined) {
        throw new Error("Could not find a uml property with id " + propertyId + " in uml type " + classToChange.name + " (id = " + classToChange.id + ")");
    }

    const errorMessage = isValidPropertyName(classToChange, newPropertySpec)
    const changedProperty: EditorUmlClassProperty = {
        id: propertyToChange.id,
        propertySpec: newPropertySpec
    }

    const changedClass: EditorUmlType = {
        id: classToChange.id,
        name: classToChange.name,
        x: classToChange.x,
        y: classToChange.y,
        methods: classToChange.methods,
        properties: classToChange.properties.set(changedProperty.id, changedProperty),
        type: classToChange.type
    }

    return [
        {
            types: umlDocument.types.set(changedClass.id, changedClass),
            inheritance: umlDocument.inheritance,
            classImplementingInterface: umlDocument.classImplementingInterface,
            relations: umlDocument.relations
        },
        errorMessage
    ]
}

/**
 * Remove a method from a uml type (like class or interface)
 * @param document The EditorUmlDocument
 * @param umlTypeId The id of the uml type (like class or interface) that is the parent of the method
 * @param methodId The uml method id that should be removed
 */
export function removeMethodFromType(document: EditorUmlDocument, umlTypeId: string, methodId: string): EditorUmlDocument {
    const umlType = document.types.get(umlTypeId)
    if (umlType == null) {
        throw Error("Couldn't find UmlType with id=" + umlTypeId)
    }

    const changedType: EditorUmlType = {
        id: umlType.id,
        name: umlType.name,
        methods: umlType.methods.remove(methodId),
        properties: umlType.properties,
        type: umlType.type,
        x: umlType.x,
        y: umlType.y
    }

    return {
        types: document.types.set(changedType.id, changedType),
        inheritance: document.inheritance,
        classImplementingInterface: document.classImplementingInterface,
        relations: document.relations
    }
}



/**
 * Remove a property from a uml class
 * 
 * @param document The EditorUmlDocument
 * @param umlTypeId The id of the uml type (like class or interface) that is the parent of the method
 * @param propertyId The uml property id that should be removed
 */
export function removePropertyFromType(document: EditorUmlDocument, umlTypeId: string, propertyId: string): EditorUmlDocument {
    const umlType = document.types.get(umlTypeId)
    if (umlType == null) {
        throw Error("Couldn't find UmlType with id=" + umlTypeId)
    }

    const changedType: EditorUmlType = {
        id: umlType.id,
        name: umlType.name,
        methods: umlType.methods,
        properties: umlType.properties.remove(propertyId),
        type: umlType.type,
        x: umlType.x,
        y: umlType.y
    }

    return {
        types: document.types.set(changedType.id, changedType),
        inheritance: document.inheritance,
        classImplementingInterface: document.classImplementingInterface,
        relations: document.relations
    }
}


/**
 * Adds a EditorUmlInheritance to the uml document.
 * 
 * @param umlDocument the original umlDocument where we want to add the new className+
 * @param superClassId The id of the superClassElement of the Connection
 * @param baseClassId The id of the baseClassElement of the Connection
 * @param umlConnectionType The Type of the Connection
 */
export function addUmlInheritance(umlDocument: EditorUmlDocument, superClassId: string, subclassId: string): EditorUmlDocument {

    const superClass = umlDocument.types.get(superClassId)
    if (superClass === undefined) {
        throw new Error("Couldn't find Type with type id = " + superClassId)
    }

    const subClass = umlDocument.types.get(subclassId)
    if (subClass === undefined) {
        throw new Error("Couldn't find Type with type id = " + subClass)
    }


    const [superClassAnchorPoint, subClassAnchorPoint] = calcAnchorPoints(superClass.x, superClass.y, subClass.x, subClass.y)

    const newInheritance: EditorUmlInheritance = {
        id: UUID.UUID(),
        superTypeId: superClassId,
        subTypeId: subclassId,
        superTypeAnchorPoint: superClassAnchorPoint,
        subTypeAnchorPoint: subClassAnchorPoint,
        points: OrderedSet<Point>()
    };


    return {
        types: umlDocument.types,
        inheritance: umlDocument.inheritance.set(newInheritance.id, newInheritance),
        classImplementingInterface: umlDocument.classImplementingInterface,
        relations: umlDocument.relations
    };
}




/**
 * Adds a EditorUmlInheritance to the uml document.
 * 
 * @param umlDocument the original umlDocument where we want to add the new className+
 * @param superClassId The id of the superClassElement of the Connection
 * @param baseClassId The id of the baseClassElement of the Connection
 * @param umlConnectionType The Type of the Connection
 */
export function addUmlRelation(umlDocument: EditorUmlDocument, startTypeId: string, endTypeId: string,
    type: UmlRelationType): EditorUmlDocument {

    const startType = umlDocument.types.get(startTypeId)
    if (startType === undefined) {
        throw new Error("Couldn't find Type with type id = " + startTypeId)
    }

    const endType = umlDocument.types.get(endTypeId)
    if (endType === undefined) {
        throw new Error("Couldn't find Type with type id = " + endTypeId)
    }


    const [startAnchorPoint, endAnchorPoint] = calcAnchorPoints(startType.x, startType.y, endType.x, endType.y)

    const newRelation: EditorUmlRelation = {
        id: UUID.UUID(),
        startTypeId: startTypeId,
        endTypeId: endTypeId,
        startTypeAnchorPoint: startAnchorPoint,
        endTypeAnchorPoint: endAnchorPoint,
        startMultiplicity: "1",
        endMultiplicity: "1",
        relationType: type,
        points: OrderedSet<Point>()
    };


    return {
        types: umlDocument.types,
        inheritance: umlDocument.inheritance,
        classImplementingInterface: umlDocument.classImplementingInterface,
        relations: umlDocument.relations.set(newRelation.id, newRelation)
    };
}



/**
 * Adds a EditorUmlInheritance to the uml document.
 * 
 * @param umlDocument the original umlDocument where we want to add the new className+
 * @param superClassId The id of the superClassElement of the Connection
 * @param baseClassId The id of the baseClassElement of the Connection
 * @param umlConnectionType The Type of the Connection
 */
export function addUmlInterfaceImplementation(umlDocument: EditorUmlDocument, interfaceId: string, classId: string): EditorUmlDocument {

    const interfaceType = umlDocument.types.get(interfaceId)
    if (interfaceType === undefined) {
        throw new Error("Couldn't find Type with type id = " + interfaceId)
    }

    const classType = umlDocument.types.get(classId)
    if (classType === undefined) {
        throw new Error("Couldn't find Type with type id = " + classId)
    }


    const [interfaceAnchorPoint, classAnchorPoint] = calcAnchorPoints(interfaceType.x, interfaceType.y, classType.x, classType.y)

    const implementation: EditorUmlInterfaceImplementation = {
        id: UUID.UUID(),
        classId: classType.id,
        interfaceId: interfaceType.id,
        interfaceAnchorPoint: interfaceAnchorPoint,
        classAnchorPoint: classAnchorPoint,
        points: OrderedSet<Point>()
    };


    return {
        types: umlDocument.types,
        inheritance: umlDocument.inheritance,
        classImplementingInterface: umlDocument.classImplementingInterface.set(implementation.id, implementation),
        relations: umlDocument.relations
    };
}



/**
 * Remove the selected elements
 * @param document the uml document
 * @param selectedElements the elements that are selected
 */
export function removeSelectedElements(document: EditorUmlDocument, selectedElements: SelectedElementsMap): EditorUmlDocument {
    var removedTypesId = Set<string>()

    const updatedTypes = document.types.withMutations(map => {
        selectedElements.forEach(element => {
            const found = map.get(element.elementId)
            if (found !== undefined) {
                removedTypesId = removedTypesId.add(found.id)
            }
            map.remove(element.elementId)
        })
    });

    const updatedInheritance = document.inheritance.withMutations(map => {
        selectedElements.forEach(element => map.remove(element.elementId))
        document.inheritance
            .filter(i => removedTypesId.contains(i.subTypeId) || removedTypesId.contains(i.superTypeId))
            .forEach(toRemove => map.remove(toRemove.id))
    })

    const updatedClassImplementsInterface = document.classImplementingInterface.withMutations(map => {
        selectedElements.forEach(element => map.remove(element.elementId))
        document.classImplementingInterface
            .filter(i => removedTypesId.contains(i.classId) || removedTypesId.contains(i.interfaceId))
            .forEach(toRemove => map.remove(toRemove.id))
    })


    const updatedRelations = document.relations.withMutations(map => {
        selectedElements.forEach(element => map.remove(element.elementId))
        document.relations
            .filter(i => removedTypesId.contains(i.startTypeId) || removedTypesId.contains(i.endTypeId))
            .forEach(toRemove => map.remove(toRemove.id))
    })


    return {
        types: updatedTypes,
        inheritance: updatedInheritance,
        classImplementingInterface: updatedClassImplementsInterface,
        relations: updatedRelations
    }
}


/**
 * Add a new Point that is used to draw a line between two types+
 * @param document The original document
 * @param umlRelationId The id of the uml Relation
 * @param x The x position of the new point
 * @param y The y position of the new point
 */
export function addPointToUmlRelation(document: EditorUmlDocument, umlRelationId: string, x: number, y: number): EditorUmlDocument {

    const relation = document.relations.get(umlRelationId)
    if (relation == null) {
        throw new Error("Couldn't find relation with id = " + umlRelationId)
    }

    const updatedRelation: EditorUmlRelation = {
        id: relation.id,
        relationType: relation.relationType,
        startTypeId: relation.startTypeId,
        endTypeId: relation.endTypeId,
        startTypeAnchorPoint: relation.startTypeAnchorPoint,
        endTypeAnchorPoint: relation.endTypeAnchorPoint,
        startMultiplicity: relation.startMultiplicity,
        endMultiplicity: relation.endMultiplicity,
        points: relation.points.add({ x: x, y: y })
    }

    return {
        types: document.types,
        inheritance: document.inheritance,
        classImplementingInterface: document.classImplementingInterface,
        relations: document.relations.set(updatedRelation.id, updatedRelation)
    }
}



/**
 * Add a new Point that is used to draw a line between two types
 * @param document The original document
 * @param umlInheritanceId The id of the uml Relation
 * @param x The x position of the new point
 * @param y The y position of the new point
 */
export function addPointToUmlInheritance(document: EditorUmlDocument, umlInheritanceId: string, x: number, y: number): EditorUmlDocument {

    const inher = document.inheritance.get(umlInheritanceId)
    if (inher == null) {
        throw new Error("Couldn't find inheritance with id = " + umlInheritanceId)
    }

    const updated: EditorUmlInheritance = {
        id: inher.id,
        subTypeAnchorPoint: inher.subTypeAnchorPoint,
        subTypeId: inher.subTypeId,
        superTypeAnchorPoint: inher.superTypeAnchorPoint,
        superTypeId: inher.superTypeId,
        points: inher.points.add({ x: x, y: y })
    }

    return {
        types: document.types,
        inheritance: document.inheritance.set(umlInheritanceId, updated),
        classImplementingInterface: document.classImplementingInterface,
        relations: document.relations
    }
}

/**
 * Add a new Point that is used to draw a line between two types
 * @param document The original document
 * @param umlInheritanceId The id of the uml Relation
 * @param x The x position of the new point
 * @param y The y position of the new point
 */
export function addPointToUmlClassImplementingInterface(document: EditorUmlDocument, umlImplementingInterfaceId: string, x: number, y: number): EditorUmlDocument {

    const inher = document.classImplementingInterface.get(umlImplementingInterfaceId)
    if (inher == null) {
        throw new Error("Couldn't find classImplementsInterface with id = " + umlImplementingInterfaceId)
    }

    const updated: EditorUmlInterfaceImplementation = {
        id: inher.id,
        classAnchorPoint: inher.classAnchorPoint,
        classId: inher.classId,
        interfaceAnchorPoint: inher.interfaceAnchorPoint,
        interfaceId: inher.interfaceId,
        points: inher.points.add({ x: x, y: y })
    }

    return {
        types: document.types,
        inheritance: document.inheritance,
        classImplementingInterface: document.classImplementingInterface.set(updated.id, updated),
        relations: document.relations
    }
}


/**
 * Changes the multiplicity of an element
 * 
 * @param document the original document
 * @param umlRelationId the id of the uml relation
 * @param updatedMultiplicity the nes string of the multiplicity
 * @param startMultiplicity true if the start mulitplicity is edited. false if end multiplicity is edited (start and end are refering to the start and end points of the uml relation)
 */
export function changeMultiplicityOfUmlRelation(document: EditorUmlDocument, umlRelationId: string, updatedMultiplicity: string, startMultiplicity: boolean): EditorUmlDocument {
    const rel = document.relations.get(umlRelationId)
    if (rel == null) {
        throw new Error("Couldn't find relation with id = " + umlRelationId)
    }

    const updated: EditorUmlRelation = {
        id: rel.id,
        startTypeId: rel.startTypeId,
        endTypeId: rel.endTypeId,
        relationType: rel.relationType,
        points: rel.points,
        startTypeAnchorPoint: rel.startTypeAnchorPoint,
        endTypeAnchorPoint: rel.endTypeAnchorPoint,
        startMultiplicity: (startMultiplicity === true ? updatedMultiplicity : rel.startMultiplicity),
        endMultiplicity: (startMultiplicity === false ? updatedMultiplicity : rel.endMultiplicity),
    }

    return {
        classImplementingInterface: document.classImplementingInterface,
        inheritance: document.inheritance,
        types: document.types,
        relations: document.relations.set(updated.id, updated)
    }
}


/**
 * Moves a AnchorPoint of a certain UML connection 
 * 
 * @param document The original document
 * @param umlRelationId The relation id
 * @param newAnchorPointPositionPercentage The AnchorpPoint's new position as percentage value
 * @param editingStartAnchorPoint true, if the "start" AnchorPoint of a connection is moved; false if "end" AnchorPoint of a connection is moved
 */
export function moveConnectionAnchorPoint(document: EditorUmlDocument, umlRelationId: string, newAnchorPointPositionPercentage: number, editingStartAnchorPoint: boolean): EditorUmlDocument {

    //
    // Anchorpoint is maybe a relation
    //
    const rel = document.relations.get(umlRelationId)
    if (rel != null) {

        let classAnchorPoint: AnchorPoint = rel.startTypeAnchorPoint
        let interfaceAnchorPoint: AnchorPoint = rel.endTypeAnchorPoint

        if (editingStartAnchorPoint === true) {
            classAnchorPoint = {
                percentage: newAnchorPointPositionPercentage
            }
        } else {
            interfaceAnchorPoint = {
                percentage: newAnchorPointPositionPercentage
            }
        }

        const updated: EditorUmlRelation = {
            id: rel.id,
            startTypeId: rel.startTypeId,
            endTypeId: rel.endTypeId,
            relationType: rel.relationType,
            points: rel.points,
            startTypeAnchorPoint: classAnchorPoint,
            endTypeAnchorPoint: interfaceAnchorPoint,
            startMultiplicity: rel.startMultiplicity,
            endMultiplicity: rel.endMultiplicity
        }

        return {
            classImplementingInterface: document.classImplementingInterface,
            inheritance: document.inheritance,
            types: document.types,
            relations: document.relations.set(updated.id, updated)
        }

    } else {
        // 
        // AnchorPoint is maybe part of inheritance
        //

        const inh = document.inheritance.get(umlRelationId)
        if (inh != null) {

            let subTypeAnchorPoint: AnchorPoint = inh.subTypeAnchorPoint
            let superTypeAnchorPoint: AnchorPoint = inh.superTypeAnchorPoint

            if (editingStartAnchorPoint === true) {
                subTypeAnchorPoint = {
                    percentage: newAnchorPointPositionPercentage
                }
            } else {
                superTypeAnchorPoint = {
                    percentage: newAnchorPointPositionPercentage
                }
            }

            const updated: EditorUmlInheritance = {
                id: inh.id,
                points: inh.points,
                subTypeAnchorPoint: subTypeAnchorPoint,
                subTypeId: inh.subTypeId,
                superTypeAnchorPoint: superTypeAnchorPoint,
                superTypeId: inh.superTypeId
            }

            return {
                classImplementingInterface: document.classImplementingInterface,
                inheritance: document.inheritance.set(updated.id, updated),
                types: document.types,
                relations: document.relations
            }

        } else {

            // 
            // Anchorpoint is maybe part of class implementing interface relation
            //

            const impl = document.classImplementingInterface.get(umlRelationId)
            if (impl != null) {

                let classAnchorPoint: AnchorPoint = impl.classAnchorPoint
                let interfaceAnchorPoint: AnchorPoint = impl.interfaceAnchorPoint

                if (editingStartAnchorPoint === true) {
                    classAnchorPoint = {
                        percentage: newAnchorPointPositionPercentage
                    }
                } else {
                    interfaceAnchorPoint = {
                        percentage: newAnchorPointPositionPercentage
                    }
                }

                const updated: EditorUmlInterfaceImplementation = {
                    id: impl.id,
                    points: impl.points,
                    classAnchorPoint: classAnchorPoint,
                    classId: impl.classId,
                    interfaceAnchorPoint: interfaceAnchorPoint,
                    interfaceId: impl.interfaceId
                }

                return {
                    classImplementingInterface: document.classImplementingInterface.set(updated.id, updated),
                    inheritance: document.inheritance,
                    types: document.types,
                    relations: document.relations
                }

            }
            else {
                throw new Error("Couldn't find any relation, inheritance or class implements an interface connection with id = " + umlRelationId)
            }

        }

    }
}
