import { EditorUmlDocument } from './EditorUmlDocument'
import { EditorUmlType, UmlType } from './EditorUmlClass'
import { EditorUmlClassProperty } from './EditorUmlClassProperty'
import { EditorUmlInheritance } from './EditorUmlInheritance'
import { EditorUmlInterfaceImplementation } from './EditorUmlInterfaceImplementation'
import { EditorUmlRelation, relationTypeToString } from './EditorUmlRelation'
import { EditorUmlTypeMethod } from './EditorUmlTypeMethod'
import { AnchorPoint } from './AnchorPoint'
import { Map, OrderedMap } from "immutable"

const VISIBILITY_PUBLIC = "PUBLIC";
const VISIBILITY_PRIVATE = "PRIVATE";


/**
 * Converts a json object to EditorUmlDocument
 * @param json the json object representation of the whole uml document
 */
export function umlDocumentToJson(doc: EditorUmlDocument): any {
    const interfaces = doc.types.filter(t => t.type === UmlType.INTERFACE)
    const classes = doc.types.filter(t => t.type === UmlType.CLASS)
    const classInheritance = doc.inheritance.filter(i => doc.types.get(i.subTypeId).type === UmlType.CLASS)
    const interfaceInheritance = doc.inheritance.filter(i => doc.types.get(i.subTypeId).type === UmlType.INTERFACE)

    return {
        classes: classes.map(c => umlTypeToJson(c, c.type)).toArray(),
        interfaces: interfaces.map(i => umlTypeToJson(i, i.type)).toArray(),
        classInheritance: classInheritance.map(i => umlInheritanceToJson(i, doc.types)).toArray(),
        interfaceInheritance: interfaceInheritance.map(i => umlInheritanceToJson(i, doc.types)).toArray(),
        relations: doc.relations.map(r => umlRelationToJson(r, doc.types)).toArray(),
        classImplementingInterface: doc.classImplementingInterface.map(i => umlClassImplementinInterfaceToJson(i, doc.types)).toArray()
    }
}

/**
 * Converts a json object to an EditorUmlType
 * @param json the json object representation of the uml type (either interface or class)
 */
function umlTypeToJson(t: EditorUmlType, umlType: UmlType): any {
    if (umlType === UmlType.CLASS) {
        return {
            name: t.name,
            properties: t.properties.map(p => umlPropertyToJson(p)).toArray(),
            methods: t.methods.map(m => umlMethodToJson(m)).toArray(),
            meta: {
                id: t.id,
                x: t.x,
                y: t.y
            }
        }

    } else {
        return {
            name: t.name,
            methods: t.methods.map(m => umlMethodToJson(m)).toArray(),
            meta: {
                id: t.id,
                x: t.x,
                y: t.y
            }
        }

    }
}

/**
 * Converts the string value to a uml visibilityl notation "+" for public "-" for private
 * 
 * @param visibility Converts the strign value to a uml notation representation
 */
function getVisibilityModifier(visibility: string): string {
    return visibility === "+" ? VISIBILITY_PUBLIC : VISIBILITY_PRIVATE;
}

/**
 * Serializes a EditorUmlTypeMethod to json
 * @param m the uml method
 */
function umlMethodToJson(m: EditorUmlTypeMethod): any {

    const trimmed = m.methodSepcification.trim();
    const visibility = getVisibilityModifierFromFirstCharacter(trimmed);
    const paranthesisStartIndex = trimmed.indexOf("(");
    const parenthesisEndIndex = trimmed.indexOf(")");
    const returnTypeIndex = trimmed.lastIndexOf(":");

    const name = trimmed.substr(1, paranthesisStartIndex - 1).trim();

    let returnType = ""
    if (returnTypeIndex < parenthesisEndIndex) {
        returnType = ""
    } else {
        returnType = trimmed.substr(returnTypeIndex + 1).trim();
    }

    let parameters = [];
    if (paranthesisStartIndex + 1 == parenthesisEndIndex) {
        parameters = [];
    } else {
        const parDef = trimmed.substr(paranthesisStartIndex + 1, parenthesisEndIndex - (paranthesisStartIndex + 1)).trim();
        if (parDef.length == 0) {
            parameters = []
        } else {
            const paramsPart = parDef.split(",");
            paramsPart.forEach(p => {
                const pDef = p.trim().split(":");
                if (pDef.length != 2) {
                    // TODO error
                } else {
                    parameters.push({ name: pDef[0].trim(), type: pDef[1].trim() });
                }
            });
        }
    }


    return {
        name: name,
        visibilityModifier: visibility,
        returnType: returnType,
        parameters: parameters,
        meta: {
            id: m.id
        }
    };
}

function getVisibilityModifierFromFirstCharacter(trimmed: string): string {
    let visibility = VISIBILITY_PRIVATE
    if (trimmed.charAt(0) === '+' || trimmed.charAt(0) === '-') {
        visibility = getVisibilityModifier(trimmed.charAt(0))
    }
    // TODO else it is an error?
    return visibility;
}

/**
 * Converts a EditorUmlClassProperty to a json representation
 * @param p the propery to serialize to json
 */
function umlPropertyToJson(p: EditorUmlClassProperty): any {
    const trimmed = p.propertySpec.trim();
    const parts = trimmed.split(":")
    let name, type;
    if (parts.length != 2) {
        // TODO error
    } else {
        name = parts[0].substr(1).trim();
        type = parts[1].trim();
    }

    return {
        visibilityModifier: getVisibilityModifierFromFirstCharacter(trimmed),
        name: name,
        type: type,
        meta: {
            id: p.id
        }
    }
}

function umlInheritanceToJson(i: EditorUmlInheritance, types: Map<string, EditorUmlType>) {
    return {
        subclassName: types.get(i.subTypeId).name,
        baseclassName: types.get(i.superTypeId).name,
        meta: {
            id: i.id,
            subTypeId: i.subTypeId,
            superTypeId: i.superTypeId,
            subAnchorPoint: i.subTypeAnchorPoint,
            superAnchorPoint: i.superTypeAnchorPoint,
            points: i.points.toJS()
        }

    }
}

function umlClassImplementinInterfaceToJson(i: EditorUmlInterfaceImplementation, types: Map<string, EditorUmlType>) {
    return {
        className: types.get(i.classId).name,
        interfaceName: types.get(i.interfaceId).name,
        meta: {
            id: i.id,
            classAnchorPoint: i.classAnchorPoint,
            interfaceAnchorPoint: i.interfaceAnchorPoint,
            classId: i.classId,
            interfaceId: i.interfaceId,
            points: i.points.toJS()
        }
    }
}

function umlRelationToJson(r: EditorUmlRelation, types: Map<string, EditorUmlType>) {
    return {
        relationType: relationTypeToString(r.relationType),
        startTypeName: types.get(r.startTypeId).name,
        endTypeName: types.get(r.endTypeId).name,
        startMultiplicity: r.startMultiplicity,
        endMultiplicity: r.endMultiplicity,
        meta: {
            id: r.id,
            startTypeId: r.startTypeId,
            endTypeId: r.endTypeId,
            startTypeAnchorPoint: r.startTypeAnchorPoint,
            endTypeAnchorPoint: r.endTypeAnchorPoint,
            points: r.points.toJS()

        }
    }
}
