import { ViewStateReducer, logReducer } from "../reducer/ViewStateReducer"
import * as Rx from "@reactivex/rxjs";
import { Intent } from "../intent/Intent"
import { StartAddingUmlTypeIntent, StartSelectingElementsIntent, FirstElementSelectedIntent, SecondElementSelectedIntent } from "../intent/InteractionTransactionIntent"
import { ViewState } from "../ViewState"
import { AddTypeInteractionTransaction, SelectingTwoElementsInteractionTransaction, NoInteractionTransactionRunning } from "../interaction/InteractionTransaction"
import { selectMenuItem } from "../MenuItem"
import { noSelectedElements } from "../SelectedElement"
import { MenuItemType } from "../MenuItem"
import { umlTypeToHumanString, UmlType } from "../model/EditorUmlClass"
import {
    InfoMessage, InfoMessageType, deleteBigInfoMessages,
    INFO_MESSAGE_CLICK_ON_CANVAS_TO_ADD_TYPE,
    INFO_MESSAGE_ERROR_INHERITANCE_SAME_TYPE_SELECTED,
    INFO_MESSAGE_SELECT_TYPE_FOR_INHERITANCE,
    INFO_MESSAGE_FOR_INTERFACE_IMPLEMENTATION,
    INFO_MESSAGE_ERROR_IMPLEMENTING_INTERFACE_SAME_TYPE_SELECTED,
    INFO_MESSAGE_SELECT_TYPE_FOR_RELATION
} from "../InfoMessage"



/**
 * Reduces the state where the user has clicked on a menu item to add a UML Type.
 * Next he wants to click on the canvas to select the place where he wants to add an element
 * @param intent$ The intent observable stream
 */
export function menuAddUmlTypeClickedReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof StartAddingUmlTypeIntent)
        .map(intent => intent as StartAddingUmlTypeIntent)
        .do(i => logReducer(log, "menuAddUmlTypeClickedReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {

            const previousSelected = oldState.menuItems.find(item => item.selected);
            if (previousSelected !== undefined && previousSelected.type === intent.menuItemType) {
                // Nothing has been changed because clicked on menu item that has been clicked before --> memory optimization
                return oldState;
            }

            return {
                menuItems: selectMenuItem(intent.menuItemType),
                umlDocument: oldState.umlDocument,
                currentInteractionTransaction: new AddTypeInteractionTransaction(intent.umlType),
                errors: oldState.errors,
                selectedElements: noSelectedElements(oldState.selectedElements),
                infoMessages: deleteBigInfoMessages(oldState.infoMessages).set(INFO_MESSAGE_CLICK_ON_CANVAS_TO_ADD_TYPE, {
                    id: INFO_MESSAGE_CLICK_ON_CANVAS_TO_ADD_TYPE,
                    type: InfoMessageType.BIG_INFO,
                    message: "Click on the canvas to create a new " + umlTypeToHumanString(intent.umlType)
                }),
            };
        }));
}



/**
 * Reduces the state where the user has clicked on a menu item to add a UML Type.
 * Next he wants to click on the canvas to select the place where he wants to add an element
 * @param intent$ The intent observable stream
 */
export function menuInheritanceClickedReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof StartSelectingElementsIntent
            && intent.menuItemType === MenuItemType.INHERITANCE
        )
        .map(intent => intent as StartSelectingElementsIntent)
        .do(i => logReducer(log, "menuInheritanceClickedReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {

            const previousSelected = oldState.menuItems.find(item => item.selected);
            if (previousSelected !== undefined && previousSelected.type === intent.menuItemType) {
                // Nothing has been changed because clicked on menu item that has been clicked before --> memory optimization
                return oldState;
            }

            return {
                menuItems: selectMenuItem(intent.menuItemType),
                umlDocument: oldState.umlDocument,
                currentInteractionTransaction: new SelectingTwoElementsInteractionTransaction(intent.menuItemType, null),
                errors: oldState.errors,
                selectedElements: noSelectedElements(oldState.selectedElements),
                infoMessages: deleteBigInfoMessages(oldState.infoMessages).set(INFO_MESSAGE_SELECT_TYPE_FOR_INHERITANCE, {
                    id: INFO_MESSAGE_SELECT_TYPE_FOR_INHERITANCE,
                    type: InfoMessageType.BIG_INFO,
                    message: "Click on a " + umlTypeToHumanString(UmlType.CLASS) + " or " + umlTypeToHumanString(UmlType.INTERFACE)
                })
            };
        }));
}


/**
 * Reduces the state where the user has clicked on a menu item to add a UML Type.
 * Next he wants to click on the canvas to select the place where he wants to add an element
 * @param intent$ The intent observable stream
 */
export function transactionSelectedFirstTypeForInheritanceReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof FirstElementSelectedIntent
            && intent.menuItemType === MenuItemType.INHERITANCE)
        .map(intent => intent as FirstElementSelectedIntent)
        .do(i => logReducer(log, "transactionSelectedFirstTypeForInheritanceReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {

            const umlType = oldState.umlDocument.types.get(intent.firstSelectedElementId);
            if (umlType === undefined) {
                throw new Error("Couldn't find a uml type with the given id = " + umlType.id);
            }

            // TODO highlight possible choises
            return {
                menuItems: selectMenuItem(intent.menuItemType),
                umlDocument: oldState.umlDocument,
                currentInteractionTransaction: new SelectingTwoElementsInteractionTransaction(intent.menuItemType, intent.firstSelectedElementId),
                errors: oldState.errors,
                selectedElements: noSelectedElements(oldState.selectedElements),
                infoMessages: deleteBigInfoMessages(oldState.infoMessages).set(INFO_MESSAGE_SELECT_TYPE_FOR_INHERITANCE, {
                    id: INFO_MESSAGE_SELECT_TYPE_FOR_INHERITANCE,
                    type: InfoMessageType.BIG_INFO,
                    message: "Click on the base " + umlTypeToHumanString(umlType.type) + " '" + umlType.name + "' should inherit from"
                })
            };
        }));
}

/**
 * Reduces the state where the user has clicked on a menu item to add a UML Type.
 * Next he wants to click on the canvas to select the place where he wants to add an element
 * @param intent$ The intent observable stream
 */
export function menuImplementsInterfaceClickedReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof StartSelectingElementsIntent
            && intent.menuItemType === MenuItemType.IMPLEMENT_INTERFACE
        )
        .map(intent => intent as StartSelectingElementsIntent)
        .do(i => logReducer(log, "menuImplementsInterfaceClickedReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {

            const previousSelected = oldState.menuItems.find(item => item.selected);
            if (previousSelected !== undefined && previousSelected.type === intent.menuItemType) {
                // Nothing has been changed because clicked on menu item that has been clicked before --> memory optimization
                return oldState;
            }

            return {
                menuItems: selectMenuItem(intent.menuItemType),
                umlDocument: oldState.umlDocument,
                currentInteractionTransaction: new SelectingTwoElementsInteractionTransaction(intent.menuItemType, null),
                errors: oldState.errors,
                selectedElements: noSelectedElements(oldState.selectedElements),
                infoMessages: deleteBigInfoMessages(oldState.infoMessages).set(INFO_MESSAGE_SELECT_TYPE_FOR_INHERITANCE, {
                    id: INFO_MESSAGE_SELECT_TYPE_FOR_INHERITANCE,
                    type: InfoMessageType.BIG_INFO,
                    message: "Click on a " + umlTypeToHumanString(UmlType.CLASS) + " that should implement a certain " + umlTypeToHumanString(UmlType.INTERFACE)
                })
            };
        }));
}

/**
 * Reduces the state where the user has clicked on a class which should implement a certain interface.
 * Next he wants to click on the canvas to select the interface the class should implement
 * @param intent$ The intent observable stream
 */
export function transactionSelectedClassForInterfaceImplementationReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof FirstElementSelectedIntent
            && intent.menuItemType === MenuItemType.IMPLEMENT_INTERFACE)
        .map(intent => intent as FirstElementSelectedIntent)
        .do(i => logReducer(log, "transactionSelectedClassForInterfaceImplementationReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {

            const umlType = oldState.umlDocument.types.get(intent.firstSelectedElementId);
            if (umlType === undefined) {
                throw new Error("Couldn't find a uml type with the given id = " + umlType.id);
            }

            if (umlType.type !== UmlType.CLASS) {
                return {
                    menuItems: oldState.menuItems,
                    umlDocument: oldState.umlDocument,
                    currentInteractionTransaction: oldState.currentInteractionTransaction,
                    errors: oldState.errors,
                    selectedElements: noSelectedElements(oldState.selectedElements),
                    infoMessages: deleteBigInfoMessages(oldState.infoMessages).set(INFO_MESSAGE_FOR_INTERFACE_IMPLEMENTATION, {
                        id: INFO_MESSAGE_FOR_INTERFACE_IMPLEMENTATION,
                        type: InfoMessageType.BIG_INFO,
                        message: "Error: Only classes can be selected but you have clicked on an Interface!\nClick on the class that should implement an interface"
                    })
                }
            }

            // TODO highlight possible choises
            return {
                menuItems: selectMenuItem(intent.menuItemType),
                umlDocument: oldState.umlDocument,
                currentInteractionTransaction: new SelectingTwoElementsInteractionTransaction(intent.menuItemType, intent.firstSelectedElementId),
                errors: oldState.errors,
                selectedElements: noSelectedElements(oldState.selectedElements),
                infoMessages: deleteBigInfoMessages(oldState.infoMessages).set(INFO_MESSAGE_FOR_INTERFACE_IMPLEMENTATION, {
                    id: INFO_MESSAGE_FOR_INTERFACE_IMPLEMENTATION,
                    type: InfoMessageType.BIG_INFO,
                    message: "Click on the Interface that should implement by the class '" + umlType.name + "'"
                })
            };
        }));
}



/**
 * Reduces the state where the user has clicked on a menu item to add a UML Type.
 * Next he wants to click on the canvas to select the place where he wants to add an element
 * @param intent$ The intent observable stream
 */
export function menuAddUmlReactionClickedReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof StartSelectingElementsIntent
            && (
                intent.menuItemType === MenuItemType.ASSOCIATION
                || intent.menuItemType === MenuItemType.AGGREGATION
                || intent.menuItemType === MenuItemType.COMPOSITION
            )
        )
        .map(intent => intent as StartSelectingElementsIntent)
        .do(i => logReducer(log, "menuAddUmlReactionClickedReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {

            const previousSelected = oldState.menuItems.find(item => item.selected);
            if (previousSelected !== undefined && previousSelected.type === intent.menuItemType) {
                // Nothing has been changed because clicked on menu item that has been clicked before --> memory optimization
                return oldState;
            }

            return {
                menuItems: selectMenuItem(intent.menuItemType),
                umlDocument: oldState.umlDocument,
                currentInteractionTransaction: new SelectingTwoElementsInteractionTransaction(intent.menuItemType, null),
                errors: oldState.errors,
                selectedElements: noSelectedElements(oldState.selectedElements),
                infoMessages: deleteBigInfoMessages(oldState.infoMessages).set(INFO_MESSAGE_SELECT_TYPE_FOR_RELATION, {
                    id: INFO_MESSAGE_SELECT_TYPE_FOR_RELATION,
                    type: InfoMessageType.BIG_INFO,
                    message: "Where does the relation starts? Click on that Class or Interface."
                })
            };
        }));
}

/**
 * Reduces the state where the user has clicked on a type (class or interface)
 * @param intent$ The intent observable stream
 */
export function transactionSelectedStartTypeForRelationReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof FirstElementSelectedIntent
            && (
                intent.menuItemType === MenuItemType.ASSOCIATION
                || intent.menuItemType === MenuItemType.AGGREGATION
                || intent.menuItemType === MenuItemType.COMPOSITION
            ))
        .map(intent => intent as FirstElementSelectedIntent)
        .do(i => logReducer(log, "transactionSelectedStartTypeForRelationReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {

            const umlType = oldState.umlDocument.types.get(intent.firstSelectedElementId);
            if (umlType === undefined) {
                throw new Error("Couldn't find a uml type with the given id = " + umlType.id);
            }

            // TODO highlight possible choises
            return {
                menuItems: selectMenuItem(intent.menuItemType),
                umlDocument: oldState.umlDocument,
                currentInteractionTransaction: new SelectingTwoElementsInteractionTransaction(intent.menuItemType, intent.firstSelectedElementId),
                errors: oldState.errors,
                selectedElements: noSelectedElements(oldState.selectedElements),
                infoMessages: deleteBigInfoMessages(oldState.infoMessages).set(INFO_MESSAGE_SELECT_TYPE_FOR_RELATION, {
                    id: INFO_MESSAGE_SELECT_TYPE_FOR_RELATION,
                    type: InfoMessageType.BIG_INFO,
                    message: "Where does the relation ends? Click on that class or interface'"
                })
            };
        }));
}
