import { ViewStateReducer, logReducer } from "../reducer/ViewStateReducer"
import * as Rx from "@reactivex/rxjs";
import { Intent } from "../intent/Intent"
import {
    ClickOnUmlCanvasIntent, SelectUmlTypeIntent, SelectUmlInterfaceImplementationIntent,
    SelectUmlTypeAndMethodIntent, SelectUmlTypeAndPropertyIntent, DeleteSelectedElementsIntent,
    SelectUmlInheritanceIntent, SelectUmlRelationIntent, MoveSelectedElementsIntent,
    StartEditingUmlRelationTextFieldIntent, CancelEditingUmlRelationTextFieldIntent,
    ChangeUmlRelationTextFieldIntent, UmlRelationTextFieldType, CancelEverythingIntent
} from "../intent/UmlCanvasIntents"
import { ViewState } from "../ViewState"
import {
    UmlTypeSelected, SelectedElement, MethodSelectedElement,
    PropertySelectedElement, noSelectedElements, setSelected, setMultipleSelected,
    getSelectedPropertyOrMethod, SelectUmlRelation
} from "../SelectedElement"
import { removeError, addOrRemoveError } from "../ErrorMessage"
import { isNoInteractionRunning, NoInteractionTransactionRunning } from "../interaction/InteractionTransaction"
import {
    removeMethodFromType, removePropertyFromType, removeSelectedElements,
    changeMultiplicityOfUmlRelation, isValidMultiplicity, umlRelationMultiplicityId
} from "../model/EditorUmlDocument"
import { UmlType, EditorUmlType } from "../model/EditorUmlClass"
import { unselectedItems } from "../MenuItem"
import { newInfoMessageMap } from "../InfoMessage"


/**
 * Creates a ViewStateReducer that deselcts all selected items and reduces its state
 * @param intent$ The original stream of intents
 */
export function deselectAllOnUmlCanvasClickReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof ClickOnUmlCanvasIntent
            && isNoInteractionRunning(intent.currentInteraction))
        .map(intent => intent as ClickOnUmlCanvasIntent)
        .do(i => logReducer(log, "deselectAllOnUmlCanvasClickReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {

            // Memory optimization. Nothing has changed
            if (oldState.selectedElements.isEmpty())
                return oldState;

            return {
                menuItems: oldState.menuItems,
                umlDocument: oldState.umlDocument,
                currentInteractionTransaction: oldState.currentInteractionTransaction,
                errors: oldState.errors,
                infoMessages: oldState.infoMessages,
                selectedElements: noSelectedElements(oldState.selectedElements)
            };
        }));
}


/**
 * Creates a ViewStateReducer that deselcts all selected items, unselects menu, removes info messages and reduces its state
 * @param intent$ The original stream of intents
 */
export function cancelEverythingReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof CancelEverythingIntent)
        .map(intent => intent as CancelEverythingIntent)
        .do(i => logReducer(log, "cancelEverythingReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {

            return {
                menuItems: unselectedItems(),
                umlDocument: oldState.umlDocument,
                currentInteractionTransaction: NoInteractionTransactionRunning.getInstance(),
                errors: oldState.errors,
                infoMessages: newInfoMessageMap(),
                selectedElements: noSelectedElements(oldState.selectedElements)
            };
        }));
}





/**
 * State reducer that reacts on selecting a single UmlType by clicking on it 
 */
export function selectUmlTypeOnUmCanvasReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof SelectUmlTypeIntent)
        .map(intent => intent as SelectUmlTypeIntent)
        .do(i => logReducer(log, "selectUmlTypeOnUmCanvasReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {
            return {
                menuItems: oldState.menuItems,
                umlDocument: oldState.umlDocument,
                currentInteractionTransaction: oldState.currentInteractionTransaction,
                errors: oldState.errors,
                infoMessages: oldState.infoMessages,
                selectedElements: setSelected(new UmlTypeSelected(intent.umlType.id, false))
            };
        }));
}

/**
 * State reducer that reacts on selecting a single UmlType Property by clicking on it 
 */
export function selectUmlTypeAndPropertyOnUmCanvasReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof SelectUmlTypeAndPropertyIntent)
        .map(intent => intent as SelectUmlTypeAndPropertyIntent)
        .do(i => logReducer(log, "selectUmlTypeAndPropertyOnUmCanvasReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {
            return {
                menuItems: oldState.menuItems,
                umlDocument: oldState.umlDocument,
                currentInteractionTransaction: oldState.currentInteractionTransaction,
                errors: oldState.errors,
                infoMessages: oldState.infoMessages,
                selectedElements: setMultipleSelected(
                    new UmlTypeSelected(intent.umlType.id, false),
                    new PropertySelectedElement(intent.umlPropertyId, intent.umlType.id, false))
            };
        }));
}


/**
 * State reducer that reacts on selecting a single UmlType Property by clicking on it 
 */
export function selectUmlTypeAndMethodOnUmCanvasReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof SelectUmlTypeAndMethodIntent)
        .map(intent => intent as SelectUmlTypeAndMethodIntent)
        .do(i => logReducer(log, "selectUmlTypeAndMethodOnUmCanvasReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {
            return {
                menuItems: oldState.menuItems,
                umlDocument: oldState.umlDocument,
                currentInteractionTransaction: oldState.currentInteractionTransaction,
                errors: oldState.errors,
                infoMessages: oldState.infoMessages,
                selectedElements: setMultipleSelected(
                    new UmlTypeSelected(intent.umlType.id, false),
                    new MethodSelectedElement(intent.umlMethodId, intent.umlType.id, false))
            };
        }));
}







/**
 * State reducer that reacts on selecting a single UmlType by clicking on it 
 */
export function deleteSelectedElementsReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof DeleteSelectedElementsIntent)
        .map(intent => intent as DeleteSelectedElementsIntent)
        .do(i => logReducer(log, "deleteSelectedElementsReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {

            // Nothing changes, so memory optimization
            if (oldState.selectedElements.isEmpty()) {
                return oldState;
            }

            let document = oldState.umlDocument
            let selections = oldState.selectedElements
            let errors = oldState.errors
            const selectedPropertyOrMethod = getSelectedPropertyOrMethod(oldState.selectedElements);
            if (selectedPropertyOrMethod instanceof MethodSelectedElement) {
                document = removeMethodFromType(oldState.umlDocument, selectedPropertyOrMethod.umlTypeId, selectedPropertyOrMethod.elementId)
                selections = setSelected(new UmlTypeSelected(selectedPropertyOrMethod.umlTypeId, false))
                errors = removeError(oldState.errors, selectedPropertyOrMethod.elementId)
            } else if (selectedPropertyOrMethod instanceof PropertySelectedElement) {
                document = removePropertyFromType(oldState.umlDocument, selectedPropertyOrMethod.umlTypeId, selectedPropertyOrMethod.elementId)
                selections = setSelected(new UmlTypeSelected(selectedPropertyOrMethod.umlTypeId, false))
                errors = removeError(oldState.errors, selectedPropertyOrMethod.elementId)
            } else {
                errors = oldState.errors.withMutations(map => {
                    oldState.selectedElements.forEach(selected => {
                        map.remove(selected.elementId)
                        // 
                        // Remove Type objects
                        //
                        const type = oldState.umlDocument.types.get(selected.elementId)
                        if (type !== undefined) {
                            type.methods.forEach(m => map.remove(m.id))
                            if (type.type === UmlType.CLASS) {
                                type.properties.forEach(p => map.remove(p.id))
                            }

                            // Remove error messages of relations that are depending on the deleted type
                            oldState.umlDocument.relations.filter(rel => rel.startTypeId === type.id || rel.endTypeId === type.id)
                                .forEach(rel => {
                                    map.remove(umlRelationMultiplicityId(rel.id, true))
                                    map.remove(umlRelationMultiplicityId(rel.id, false))
                                    map.remove(rel.id)
                                })

                            // Remove error messages of inheritance that are depending on the deleted type
                            oldState.umlDocument.inheritance.filter(inh => inh.subTypeId === type.id || inh.superTypeId === type.id)
                                .forEach(inh => map.remove(inh.id))

                            // Remove error messages of interface implementation that are depending on the deleted type
                            oldState.umlDocument.classImplementingInterface.filter(impl => impl.classId === type.id || impl.interfaceId === type.id)
                                .forEach(impl => map.remove(impl.id))
                        }

                        //
                        // Remove Relations
                        //
                        const relation = oldState.umlDocument.relations.get(selected.elementId)
                        if (relation !== undefined) {
                            map.remove(relation.id)
                            map.remove(umlRelationMultiplicityId(relation.id, true))
                            map.remove(umlRelationMultiplicityId(relation.id, false))
                        }

                        //
                        // Remove Inheritance
                        //
                        const inheritance = oldState.umlDocument.inheritance.get(selected.elementId)
                        if (inheritance !== undefined) {
                            map.remove(inheritance.id)
                        }

                        //
                        // Remove class implements interface
                        //
                        const classImplInterface = oldState.umlDocument.classImplementingInterface.get(selected.elementId)
                        if (classImplInterface !== undefined) {
                            map.remove(classImplInterface.id)
                        }
                    })
                })
                selections = noSelectedElements(oldState.selectedElements)
                document = removeSelectedElements(oldState.umlDocument, oldState.selectedElements)
            }

            return {
                menuItems: oldState.menuItems,
                umlDocument: document,
                currentInteractionTransaction: oldState.currentInteractionTransaction,
                errors: errors,
                infoMessages: oldState.infoMessages,
                selectedElements: selections
            };
        }));
}


/**
 * State reducer that reacts on selecting a single UmlType by clicking on it 
 */
export function selectUmlInheritanceOnCanvasReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof SelectUmlInheritanceIntent)
        .map(intent => intent as SelectUmlInheritanceIntent)
        .do(i => logReducer(log, "selectUmlInheritanceOnCanvasReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {
            return {
                menuItems: oldState.menuItems,
                umlDocument: oldState.umlDocument,
                currentInteractionTransaction: oldState.currentInteractionTransaction,
                errors: oldState.errors,
                infoMessages: oldState.infoMessages,
                selectedElements: setSelected(new SelectedElement(intent.umlInheritanceId))
            };
        }));
}


/**
 * State reducer that reacts on selecting a single inheritance implementation by clicking on it 
 */
export function selectUmlImplementInterfacenReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof SelectUmlInterfaceImplementationIntent)
        .map(intent => intent as SelectUmlInterfaceImplementationIntent)
        .do(i => logReducer(log, "selectUmlImplementInterfacenReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {
            return {
                menuItems: oldState.menuItems,
                umlDocument: oldState.umlDocument,
                currentInteractionTransaction: oldState.currentInteractionTransaction,
                errors: oldState.errors,
                infoMessages: oldState.infoMessages,
                selectedElements: setSelected(new SelectedElement(intent.umlInterfaceImplementation))
            };
        }));
}


/**
 * State reducer that reacts on selecting a single UmlRelation by clicking on it 
 */
export function selectUmlRelationReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof SelectUmlRelationIntent)
        .map(intent => intent as SelectUmlRelationIntent)
        .do(i => logReducer(log, "selectUmlRelationReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {
            return {
                menuItems: oldState.menuItems,
                umlDocument: oldState.umlDocument,
                currentInteractionTransaction: oldState.currentInteractionTransaction,
                errors: oldState.errors,
                infoMessages: oldState.infoMessages,
                selectedElements: setSelected(new SelectUmlRelation(intent.umlRelationId, false, false))
            };
        }));
}


/**
 * The intent to add a uml relation like association, aggregation and composition
 * @param intent$ the stream of intents
 * @param log The log intent
 */
export function startEditinUmlRelationTextFieldReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof StartEditingUmlRelationTextFieldIntent)
        .map(intent => intent as StartEditingUmlRelationTextFieldIntent)
        .do(i => logReducer(log, "startEditinUmlRelationTextFieldReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {

            var selectedElement: SelectUmlRelation

            switch (intent.relationField) {
                case UmlRelationTextFieldType.START_MULTIPLICITY:
                    selectedElement = new SelectUmlRelation(intent.umlRelationId, true, false);
                    break
                case UmlRelationTextFieldType.END_MULTIPLICITY:
                    selectedElement = new SelectUmlRelation(intent.umlRelationId, false, true);
                    break
                default:
                    throw new Error("Don't know which input field: " + intent.relationField)
            }

            return {
                menuItems: oldState.menuItems,
                umlDocument: oldState.umlDocument,
                currentInteractionTransaction: oldState.currentInteractionTransaction,
                errors: oldState.errors,
                selectedElements: setSelected(selectedElement),
                infoMessages: oldState.infoMessages
            }
        }));
}


/**
 * The intent to add a uml relation like association, aggregation and composition
 * @param intent$ the stream of intents
 * @param log The log intent
 */
export function cancelEditinUmlRelationTextFieldReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof CancelEditingUmlRelationTextFieldIntent)
        .map(intent => intent as CancelEditingUmlRelationTextFieldIntent)
        .do(i => logReducer(log, "cancelEditinUmlRelationTextFieldReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {


            return {
                menuItems: oldState.menuItems,
                umlDocument: oldState.umlDocument,
                currentInteractionTransaction: oldState.currentInteractionTransaction,
                errors: oldState.errors,
                selectedElements: setSelected(new SelectUmlRelation(intent.umlRelationId, false, false)),
                infoMessages: oldState.infoMessages
            }
        }));
}


/**
 * The intent to add a uml relation like association, aggregation and composition
 * @param intent$ the stream of intents
 * @param log The log intent
 */
export function changeUmlRelationTextFieldReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof ChangeUmlRelationTextFieldIntent)
        .map(intent => intent as ChangeUmlRelationTextFieldIntent)
        .do(i => logReducer(log, "changeUmlRelationTextFieldReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {

            var startMultiplicity = true
            switch (intent.relationField) {
                case UmlRelationTextFieldType.START_MULTIPLICITY:
                    startMultiplicity = true
                    break

                case UmlRelationTextFieldType.END_MULTIPLICITY:
                    startMultiplicity = false
                    break

                default:
                    throw new Error("Not implemented for TextFieldType " + intent.relationField)
            }



            const error = isValidMultiplicity(intent.updatedValue)
            var errors = addOrRemoveError(oldState.errors, umlRelationMultiplicityId(intent.umlRelationId, startMultiplicity), error)


            return {
                menuItems: oldState.menuItems,
                umlDocument: changeMultiplicityOfUmlRelation(oldState.umlDocument, intent.umlRelationId, intent.updatedValue, startMultiplicity),
                currentInteractionTransaction: oldState.currentInteractionTransaction,
                errors: errors,
                selectedElements: setSelected(new SelectUmlRelation(intent.umlRelationId, false, false)),
                infoMessages: oldState.infoMessages
            }
        }));
}




/**
 * State reducer that reacts on selecting a single UmlRelation by clicking on it 
 */
export function moveSelectedElementsReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof MoveSelectedElementsIntent)
        .map(intent => intent as MoveSelectedElementsIntent)
        .do(i => logReducer(log, "moveSelectedElementsReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {

            if (oldState.selectedElements.isEmpty()) {
                return oldState;
            }

            const elementsToMove = oldState.selectedElements.filter(element => element instanceof UmlTypeSelected)
            if (elementsToMove.isEmpty()) {
                return oldState;
            }


            const updatedTypes = oldState.umlDocument.types.withMutations(map => {
                elementsToMove.forEach(element => {
                    const umlType = oldState.umlDocument.types.get(element.elementId)
                    if (umlType == null) {
                        throw new Error("Could not find uml type with id=" + element.elementId)
                    }

                    // TODO check if that is too memory intensive
                    const updated: EditorUmlType = {
                        id: umlType.id,
                        methods: umlType.methods,
                        properties: umlType.properties,
                        name: umlType.name,
                        type: umlType.type,
                        x: intent.newX,
                        y: intent.newY
                    }

                    map.set(updated.id, updated)
                })
            })

            // Memory optimization
            intent.recycle()

            return {
                menuItems: oldState.menuItems,
                umlDocument: {
                    types: updatedTypes,
                    classImplementingInterface: oldState.umlDocument.classImplementingInterface,
                    inheritance: oldState.umlDocument.inheritance,
                    relations: oldState.umlDocument.relations
                },
                currentInteractionTransaction: oldState.currentInteractionTransaction,
                errors: oldState.errors,
                infoMessages: oldState.infoMessages,
                selectedElements: oldState.selectedElements
            };
        }));
}
