import { ViewStateReducer, logReducer } from "../reducer/ViewStateReducer"
import * as Rx from "@reactivex/rxjs";
import { Intent } from "../intent/Intent"
import { selectMenuItem, unselectedItems } from "../MenuItem"
import { noSelectedElements } from "../SelectedElement"
import { MenuItemType } from "../MenuItem"
import { umlTypeToHumanString, UmlType } from "../model/EditorUmlClass"
import { InfoMessage, InfoMessageType, deleteBigInfoMessages, INFO_MESSAGE_ERROR_INHERITANCE_SAME_TYPE_SELECTED, INFO_MESSAGE_SELECT_TYPE_FOR_INHERITANCE } from "../InfoMessage"
import { SecondElementSelectedIntent } from "../intent/InteractionTransactionIntent"
import { ViewState } from "../ViewState"
import { NoInteractionTransactionRunning } from "../interaction/InteractionTransaction"
import { ERROR_DISPLAY_TIME } from "../ErrorMessage"
import { addUmlInheritance } from "../model/EditorUmlDocument"


/**
 * The intent to add inheritance
 * @param intent$ the stream of intents
 * @param log The log intent
 */
export function addInheritanceReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof SecondElementSelectedIntent
            && intent.menuItemType === MenuItemType.INHERITANCE)
        .map(intent => intent as SecondElementSelectedIntent)
        .do(i => logReducer(log, "addInheritanceReducer"))
        .switchMap(intent => {


            const successfulReducerFunction = (oldState: ViewState): ViewState => {

                const umlType1 = oldState.umlDocument.types.get(intent.firstSelectedElementId);
                if (umlType1 === undefined) {
                    throw new Error("Couldn't find a uml type with the given id = " + umlType1.id);
                }


                const umlType2 = oldState.umlDocument.types.get(intent.secondSelectedElementId);
                if (umlType2 === undefined) {
                    throw new Error("Couldn't find a uml type with the given id = " + umlType1.id);
                }

                // Inheritance must be of same type
                if (umlType1.type !== umlType2.type) {
                    const expected = umlTypeToHumanString(umlType1.type)
                    return {
                        menuItems: oldState.menuItems,
                        umlDocument: oldState.umlDocument,
                        currentInteractionTransaction: oldState.currentInteractionTransaction,
                        errors: oldState.errors,
                        selectedElements: noSelectedElements(oldState.selectedElements),
                        infoMessages: deleteBigInfoMessages(oldState.infoMessages)
                            .set(INFO_MESSAGE_SELECT_TYPE_FOR_INHERITANCE, {
                                id: INFO_MESSAGE_SELECT_TYPE_FOR_INHERITANCE,
                                message: "Error: A " + expected + " can only extend from another " + expected + "\nFrom which " + expected + " should '" + umlType1.name + "' inherit",
                                type: InfoMessageType.BIG_INFO
                            })
                    };
                }

                // check inheritance already exists
                const existingCount = oldState.umlDocument.inheritance
                .filter(i => i.subTypeId === umlType1.id && i.superTypeId === umlType2.id)
                .size
                if(existingCount > 0){
                     const expected = umlTypeToHumanString(umlType1.type)
                    return {
                        menuItems: oldState.menuItems,
                        umlDocument: oldState.umlDocument,
                        currentInteractionTransaction: oldState.currentInteractionTransaction,
                        errors: oldState.errors,
                        selectedElements: noSelectedElements(oldState.selectedElements),
                        infoMessages: deleteBigInfoMessages(oldState.infoMessages)
                            .set(INFO_MESSAGE_SELECT_TYPE_FOR_INHERITANCE, {
                                id: INFO_MESSAGE_SELECT_TYPE_FOR_INHERITANCE,
                                message: "Error: '"+umlType1.name+"' already inherits from '"+umlType2.id+"'. Choose another "+umlTypeToHumanString(umlType1.type),
                                type: InfoMessageType.BIG_INFO
                            })
                    };
                }

                return {
                    menuItems: unselectedItems(),
                    umlDocument: addUmlInheritance(oldState.umlDocument, intent.secondSelectedElementId, intent.firstSelectedElementId),
                    currentInteractionTransaction: NoInteractionTransactionRunning.getInstance(),
                    errors: oldState.errors,
                    selectedElements: noSelectedElements(oldState.selectedElements),
                    infoMessages: deleteBigInfoMessages(oldState.infoMessages)
                };
            } // End successfulFunc


            // Remove the error message
            const removeErrorSameTypemessageReducerFunction = (oldState: ViewState): ViewState => {
                return {
                    menuItems: oldState.menuItems,
                    umlDocument: oldState.umlDocument,
                    currentInteractionTransaction: oldState.currentInteractionTransaction,
                    errors: oldState.errors,
                    selectedElements: oldState.selectedElements,
                    infoMessages: oldState.infoMessages.delete(INFO_MESSAGE_ERROR_INHERITANCE_SAME_TYPE_SELECTED)
                };
            }

            const addErrorSameTypeReducerFunction = (oldState: ViewState): ViewState => {
                return {
                    menuItems: unselectedItems(),
                    umlDocument: oldState.umlDocument,
                    currentInteractionTransaction: NoInteractionTransactionRunning.getInstance(),
                    errors: oldState.errors,
                    selectedElements: noSelectedElements(oldState.selectedElements),
                    infoMessages: deleteBigInfoMessages(oldState.infoMessages).set(INFO_MESSAGE_ERROR_INHERITANCE_SAME_TYPE_SELECTED, {
                        id: INFO_MESSAGE_ERROR_INHERITANCE_SAME_TYPE_SELECTED,
                        message: "Error: You cannot select the same type as base type to inherit from",
                        type: InfoMessageType.BIG_INFO
                    })
                };
            }


            if (intent.firstSelectedElementId === intent.secondSelectedElementId) {
                return Rx.Observable.interval(ERROR_DISPLAY_TIME)
                    .take(1)
                    .map(ignored => removeErrorSameTypemessageReducerFunction)
                    .startWith(addErrorSameTypeReducerFunction)
            }

            return Rx.Observable.of(successfulReducerFunction)

        });  // End switchmap
}
