import { ViewStateReducer, logReducer } from "../reducer/ViewStateReducer"
import * as Rx from "@reactivex/rxjs";
import { Intent } from "../intent/Intent"
import { selectMenuItem, unselectedItems } from "../MenuItem"
import { noSelectedElements } from "../SelectedElement"
import { MenuItemType } from "../MenuItem"
import { umlTypeToHumanString, UmlType } from "../model/EditorUmlClass"
import { InfoMessage, InfoMessageType, deleteBigInfoMessages, INFO_MESSAGE_FOR_INTERFACE_IMPLEMENTATION, INFO_MESSAGE_ERROR_IMPLEMENTING_INTERFACE_SAME_TYPE_SELECTED } from "../InfoMessage"
import { SecondElementSelectedIntent } from "../intent/InteractionTransactionIntent"
import { ViewState } from "../ViewState"
import { NoInteractionTransactionRunning } from "../interaction/InteractionTransaction"
import { ERROR_DISPLAY_TIME } from "../ErrorMessage"
import { addUmlInterfaceImplementation } from "../model/EditorUmlDocument"


/**
 * The intent to add a interface implementation between a class and an interface
 * @param intent$ the stream of intents
 * @param log The log intent
 */
export function addInterfaceImplementationReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof SecondElementSelectedIntent
            && intent.menuItemType === MenuItemType.IMPLEMENT_INTERFACE)
        .map(intent => intent as SecondElementSelectedIntent)
        .do(i => logReducer(log, "addInterfaceImplementationReducer"))
        .switchMap(intent => {


            const successfulReducerFunction = (oldState: ViewState): ViewState => {

                const umlType1 = oldState.umlDocument.types.get(intent.firstSelectedElementId);
                if (umlType1 === undefined) {
                    throw new Error("Couldn't find a uml type with the given id = " + umlType1.id);
                }


                const umlType2 = oldState.umlDocument.types.get(intent.secondSelectedElementId);
                if (umlType2 === undefined) {
                    throw new Error("Couldn't find a uml type with the given id = " + umlType1.id);
                }

                // Check if it is a interface
                if (umlType2.type !== UmlType.INTERFACE) {

                    return {
                        menuItems: oldState.menuItems,
                        umlDocument: oldState.umlDocument,
                        currentInteractionTransaction: oldState.currentInteractionTransaction,
                        errors: oldState.errors,
                        selectedElements: noSelectedElements(oldState.selectedElements),
                        infoMessages: deleteBigInfoMessages(oldState.infoMessages).set(INFO_MESSAGE_FOR_INTERFACE_IMPLEMENTATION, {
                            id: INFO_MESSAGE_FOR_INTERFACE_IMPLEMENTATION,
                            message: "Error: You must select a interface by clicking on it!\nWhich Interface should be implement by the class '" + umlType1.name + "'?",
                            type: InfoMessageType.BIG_INFO
                        })
                    }
                }

                const alreadyExisting = oldState.umlDocument.classImplementingInterface.filter(impl => impl.classId === umlType1.id && impl.interfaceId == umlType2.id).size

                if (alreadyExisting > 0) {
                    return {
                        menuItems: oldState.menuItems,
                        umlDocument: oldState.umlDocument,
                        currentInteractionTransaction: oldState.currentInteractionTransaction,
                        errors: oldState.errors,
                        selectedElements: noSelectedElements(oldState.selectedElements),
                        infoMessages: deleteBigInfoMessages(oldState.infoMessages).set(INFO_MESSAGE_FOR_INTERFACE_IMPLEMENTATION, {
                            id: INFO_MESSAGE_FOR_INTERFACE_IMPLEMENTATION,
                            message: "Error: Class '" + umlType1.name + " already implements interface '" + umlType2.name + "'. Choose another interface or press ESC to cancel.",
                            type: InfoMessageType.BIG_INFO
                        })

                    }
                }

                return {
                    menuItems: unselectedItems(),
                    umlDocument: addUmlInterfaceImplementation(oldState.umlDocument, intent.secondSelectedElementId, intent.firstSelectedElementId),
                    currentInteractionTransaction: NoInteractionTransactionRunning.getInstance(),
                    errors: oldState.errors,
                    selectedElements: noSelectedElements(oldState.selectedElements),
                    infoMessages: deleteBigInfoMessages(oldState.infoMessages)
                };
            } // End successfulFunc


            // Remove the error message
            const removeErrorSameTypemessageReducerFunction = (oldState: ViewState): ViewState => {
                return {
                    menuItems: oldState.menuItems,
                    umlDocument: oldState.umlDocument,
                    currentInteractionTransaction: oldState.currentInteractionTransaction,
                    errors: oldState.errors,
                    selectedElements: oldState.selectedElements,
                    infoMessages: oldState.infoMessages.delete(INFO_MESSAGE_ERROR_IMPLEMENTING_INTERFACE_SAME_TYPE_SELECTED)
                };
            }

            const addErrorSameTypeReducerFunction = (oldState: ViewState): ViewState => {
                return {
                    menuItems: unselectedItems(),
                    umlDocument: oldState.umlDocument,
                    currentInteractionTransaction: NoInteractionTransactionRunning.getInstance(),
                    errors: oldState.errors,
                    selectedElements: noSelectedElements(oldState.selectedElements),
                    infoMessages: deleteBigInfoMessages(oldState.infoMessages).set(INFO_MESSAGE_ERROR_IMPLEMENTING_INTERFACE_SAME_TYPE_SELECTED, {
                        id: INFO_MESSAGE_ERROR_IMPLEMENTING_INTERFACE_SAME_TYPE_SELECTED,
                        message: "Error: You cannot select the same type",
                        type: InfoMessageType.BIG_INFO
                    })
                };
            }


            if (intent.firstSelectedElementId === intent.secondSelectedElementId) {
                return Rx.Observable.interval(ERROR_DISPLAY_TIME)
                    .take(1)
                    .map(ignored => removeErrorSameTypemessageReducerFunction)
                    .startWith(addErrorSameTypeReducerFunction)
            }

            return Rx.Observable.of(successfulReducerFunction)

        });  // End switchmap
}
