import { ViewStateReducer, logReducer } from "../reducer/ViewStateReducer"
import * as Rx from "@reactivex/rxjs";
import { Intent } from "../intent/Intent"
import { selectMenuItem, unselectedItems } from "../MenuItem"
import { noSelectedElements, setSelected } from "../SelectedElement"
import { MenuItemType } from "../MenuItem"
import { UmlRelationType } from "../model/EditorUmlRelation"
import { addPointToUmlRelation, addPointToUmlClassImplementingInterface, addPointToUmlInheritance } from "../model/EditorUmlDocument"
import { InfoMessage, InfoMessageType, deleteBigInfoMessages, INFO_MESSAGE_ERROR_INHERITANCE_SAME_TYPE_SELECTED, INFO_MESSAGE_SELECT_TYPE_FOR_INHERITANCE } from "../InfoMessage"
import { SecondElementSelectedIntent } from "../intent/InteractionTransactionIntent"
import {
    DoubleClickUmlConnectionIntent, StartEditingUmlRelationTextFieldIntent, CancelEditingUmlRelationTextFieldIntent,
    MoveAnchorPointIntent, StartMovingAnchorPointIntent, StopMovingAnchorPointIntent
} from "../intent/UmlCanvasIntents"
import { ViewState } from "../ViewState"
import { NoInteractionTransactionRunning, MovingAnchorPointInteractionTransaction } from "../interaction/InteractionTransaction"
import { ERROR_DISPLAY_TIME } from "../ErrorMessage"
import { addUmlRelation, moveConnectionAnchorPoint } from "../model/EditorUmlDocument"
import { UmlConnectionType } from "../../components/uml/UmlConnectionCanvas"


/**
 * The intent to add a uml relation like association, aggregation and composition
 * @param intent$ the stream of intents
 * @param log The log intent
 */
export function addRelationReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof SecondElementSelectedIntent
            && (
                intent.menuItemType === MenuItemType.AGGREGATION
                || intent.menuItemType === MenuItemType.ASSOCIATION
                || intent.menuItemType === MenuItemType.COMPOSITION
            )

        )
        .map(intent => intent as SecondElementSelectedIntent)
        .do(i => logReducer(log, "addRelationReducer"))
        .switchMap(intent => {


            const successfulReducerFunction = (oldState: ViewState): ViewState => {

                const umlType1 = oldState.umlDocument.types.get(intent.firstSelectedElementId);
                if (umlType1 === undefined) {
                    throw new Error("Couldn't find a uml type with the given id = " + umlType1.id);
                }


                const umlType2 = oldState.umlDocument.types.get(intent.secondSelectedElementId);
                if (umlType2 === undefined) {
                    throw new Error("Couldn't find a uml type with the given id = " + umlType1.id);
                }


                return {
                    menuItems: unselectedItems(),
                    umlDocument: addUmlRelation(oldState.umlDocument, intent.secondSelectedElementId, intent.firstSelectedElementId, menuTypeToRelationType(intent.menuItemType)),
                    currentInteractionTransaction: NoInteractionTransactionRunning.getInstance(),
                    errors: oldState.errors,
                    selectedElements: noSelectedElements(oldState.selectedElements),
                    infoMessages: deleteBigInfoMessages(oldState.infoMessages)
                };
            } // End successfulFunc


            return Rx.Observable.of(successfulReducerFunction)

        });  // End switchmap
}


function menuTypeToRelationType(menuType: MenuItemType) {
    switch (menuType) {
        case MenuItemType.ASSOCIATION: return UmlRelationType.ASSOCIATION
        case MenuItemType.AGGREGATION: return UmlRelationType.AGGREGATION
        case MenuItemType.COMPOSITION: return UmlRelationType.COMPOSITION
    }

    throw new Error("Unknown MenuType " + menuType)
}

export function addConnectionPointReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof DoubleClickUmlConnectionIntent)
        .map(intent => intent as DoubleClickUmlConnectionIntent)
        .do(i => logReducer(log, "addConnectionPointReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {

            var document = oldState.umlDocument
            switch (intent.connectionType) {
                case UmlConnectionType.INHERITANCE:
                    document = addPointToUmlInheritance(oldState.umlDocument, intent.umlConnectionId, intent.position.x, intent.position.y)
                    break;
                case UmlConnectionType.INTERFACE_IMPLEMENTATION:
                    document = addPointToUmlClassImplementingInterface(oldState.umlDocument, intent.umlConnectionId, intent.position.x, intent.position.y)
                    break;
                case UmlConnectionType.RELATION:
                    document = addPointToUmlRelation(oldState.umlDocument, intent.umlConnectionId, intent.position.x, intent.position.y)
                    break;
                default: throw new Error("Don't know which connection type it is. Type = " + intent.connectionType);
            }

            return {
                menuItems: unselectedItems(),
                umlDocument: document,
                currentInteractionTransaction: NoInteractionTransactionRunning.getInstance(),
                errors: oldState.errors,
                selectedElements: oldState.selectedElements,
                infoMessages: oldState.infoMessages
            }
        }));
}



/**
 * Move an AnchorPoint of a certain UML connection like Inheritance, Association, Composition, Aggregation and interface implementation
 */
export function moveConnectionAnchorPointReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof MoveAnchorPointIntent)
        .map(intent => intent as MoveAnchorPointIntent)
        .do(i => logReducer(log, "moveConnectionAnchorPointReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {

            const connectionId = intent.umlConnectionId
            const newPercentage = intent.newPercentage
            const firstAnchorPoint = intent.firstAnchorPoint

            intent.recycle()

            return {
                menuItems: oldState.menuItems,
                umlDocument: moveConnectionAnchorPoint(oldState.umlDocument, connectionId, newPercentage, firstAnchorPoint),
                currentInteractionTransaction: oldState.currentInteractionTransaction,
                errors: oldState.errors,
                selectedElements: oldState.selectedElements,
                infoMessages: oldState.infoMessages
            }
        }));
}


/**
 * Starts moving an AnchorPoint of a certain UML connection like Inheritance, Association, Composition, Aggregation and interface implementation
 */
export function startMovingConnectionAnchorPointReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof StartMovingAnchorPointIntent)
        .map(intent => intent as StartMovingAnchorPointIntent)
        .do(i => logReducer(log, "startMovingConnectionAnchorPointReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {

            return {
                menuItems: unselectedItems(),
                umlDocument: oldState.umlDocument,
                currentInteractionTransaction: new MovingAnchorPointInteractionTransaction(intent.umlConnectionId, intent.firstAnchorPoint, intent.minMouseValue, intent.maxMouseValue, intent.achsis),
                errors: oldState.errors,
                selectedElements: oldState.selectedElements,
                infoMessages: oldState.infoMessages
            }
        }));
}


/**
 * Stops moving an AnchorPoint of a certain UML connection like Inheritance, Association, Composition, Aggregation and interface implementation
 */
export function stopMovingConnectionAnchorPointReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof StopMovingAnchorPointIntent)
        .map(intent => intent as StopMovingAnchorPointIntent)
        .do(i => logReducer(log, "stopMovingConnectionAnchorPointReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {

            return {
                menuItems: oldState.menuItems,
                umlDocument: oldState.umlDocument,
                currentInteractionTransaction: NoInteractionTransactionRunning.getInstance(),
                errors: oldState.errors,
                selectedElements: oldState.selectedElements,
                infoMessages: oldState.infoMessages
            }
        }));
}
