import { ViewStateReducer, logReducer } from "../reducer/ViewStateReducer"
import * as Rx from "@reactivex/rxjs";
import { Intent } from "../intent/Intent"
import {
    RenameUmlTypeIntent, CancelRenameUmlTypeIntent, StartRenamingUmlTypeIntent, AddUmlMethodIntent,
    CancelEditingUmlMethodIntent, StartEditingUmlMethodIntent, ChangeUmlMethodIntent, AddUmlPropertyIntent,
    CancelEditingUmlPropertyIntent, StartEditingUmlPropertyIntent, ChangeUmlPropertyIntent
} from "../intent/UmlTypeIntents"
import { ViewState } from "../ViewState"
import {
    addUmlType, renameUmlType, addUmlMethod, changeUmlMethodSpecification, addUmlProperty, changeUmlProperty
} from "../model/EditorUmlDocument"
import { NoInteractionTransactionRunning } from "../interaction/InteractionTransaction"
import { ClickOnUmlCanvasIntent } from "../intent/UmlCanvasIntents"
import { AddTypeInteractionTransaction } from "../interaction/InteractionTransaction"
import { unselectedItems } from "../MenuItem"
import { deleteBigInfoMessages } from "../InfoMessage"
import { addOrRemoveError } from "../ErrorMessage"
import { UmlTypeSelected, MethodSelectedElement, PropertySelectedElement, setSelected, noSelectedElements, setMultipleSelected } from "../SelectedElement"
import { UUID } from "angular2-uuid"



/**
 * 
 * @param intent$ The intent observable stream
 */
export function addUmlTypeAtPositionReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof ClickOnUmlCanvasIntent
            && intent.currentInteraction instanceof AddTypeInteractionTransaction)
        .map(intent => intent as ClickOnUmlCanvasIntent)
        .do(i => logReducer(log, "addUmlTypeAtPositionReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {
            const interaction = intent.currentInteraction as AddTypeInteractionTransaction;
            const umlTypeId = UUID.UUID()
            const [document, errorMsg] = addUmlType(oldState.umlDocument, umlTypeId, "New", intent.x, intent.y, interaction.umlType)

            return {
                menuItems: unselectedItems(),
                umlDocument: document,
                currentInteractionTransaction: NoInteractionTransactionRunning.getInstance(),
                errors: addOrRemoveError(oldState.errors, umlTypeId, errorMsg),
                selectedElements: setSelected(new UmlTypeSelected(umlTypeId, true)),
                infoMessages: deleteBigInfoMessages(oldState.infoMessages)
            };
        }));
}

/**
 * Creates a ViewStateReducer that changes the name of a new UmlClass and reduces its state
 * @param intent$ The original stream of intents
 */
export function renameUmlTypeReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof RenameUmlTypeIntent)
        .map(intent => intent as RenameUmlTypeIntent)
        .do(i => logReducer(log, "renameUmlTypeReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {

            const [document, errorMsg] = renameUmlType(oldState.umlDocument, intent.umlType.id, intent.newName)

            return {
                menuItems: oldState.menuItems,
                umlDocument: document,
                currentInteractionTransaction: NoInteractionTransactionRunning.getInstance(),
                errors: addOrRemoveError(oldState.errors, intent.umlType.id, errorMsg),
                selectedElements: noSelectedElements(oldState.selectedElements),
                infoMessages: oldState.infoMessages
            };
        }));
}


/**
 * Creates a ViewStateReducer that changes the name of a new UmlClass and reduces its state
 * @param intent$ The original stream of intents
 */
export function cancelRenamingUmlTypeReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof CancelRenameUmlTypeIntent)
        .map(intent => intent as CancelRenameUmlTypeIntent)
        .do(i => logReducer(log, "cancelRenamingUmlClassReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {
            const umlTypeId = intent.umlType.id
            return {
                menuItems: oldState.menuItems,
                umlDocument: oldState.umlDocument,
                currentInteractionTransaction: NoInteractionTransactionRunning.getInstance(),
                errors: oldState.errors,
                infoMessages: oldState.infoMessages,
                selectedElements: oldState.selectedElements.set(umlTypeId, new UmlTypeSelected(umlTypeId, false))
            };
        }));
}

/**
 * Creates a ViewStateReducer that changes the name of a new UmlClass and reduces its state
 * @param intent$ The original stream of intents
 */
export function startRenamingUmlTypeReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof StartRenamingUmlTypeIntent)
        .map(intent => intent as StartRenamingUmlTypeIntent)
        .do(i => logReducer(log, "startRenamingUmlTypeReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {
            return {
                menuItems: oldState.menuItems,
                umlDocument: oldState.umlDocument,
                currentInteractionTransaction: NoInteractionTransactionRunning.getInstance(),
                errors: oldState.errors,
                infoMessages: oldState.infoMessages,
                selectedElements: setSelected(new UmlTypeSelected(intent.umlType.id, true))
            };
        }));
}

/**
 * Creates a ViewStateReducer to add a uml method to the a uml type
 * @param intent$ The original stream of intents
 */
export function addUmlMethodReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof AddUmlMethodIntent)
        .map(intent => intent as AddUmlMethodIntent)
        .do(i => logReducer(log, "addUmlMethodReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {
            const methodId = UUID.UUID()
            const [document, errorMsg] = addUmlMethod(oldState.umlDocument, intent.umlType.id, methodId)
            return {
                menuItems: oldState.menuItems,
                umlDocument: document,
                currentInteractionTransaction: NoInteractionTransactionRunning.getInstance(),
                errors: addOrRemoveError(oldState.errors, methodId, errorMsg),
                infoMessages: oldState.infoMessages,
                selectedElements: setMultipleSelected(
                    new UmlTypeSelected(intent.umlType.id, false),
                    new MethodSelectedElement(methodId, intent.umlType.id, true))
            };
        }));
}


/**
 * Creates a ViewStateReducer that sets a uml method into editing mode
 * @param intent$ The original stream of intents
 */
export function startEditingUmlMethodReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof StartEditingUmlMethodIntent)
        .map(intent => intent as StartEditingUmlMethodIntent)
        .do(i => logReducer(log, "startEditingUmlMethodReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {
            return {
                menuItems: oldState.menuItems,
                umlDocument: oldState.umlDocument,
                currentInteractionTransaction: NoInteractionTransactionRunning.getInstance(),
                errors: oldState.errors,
                infoMessages: oldState.infoMessages,
                selectedElements: setMultipleSelected(
                    new UmlTypeSelected(intent.umlType.id, false),
                    new MethodSelectedElement(intent.methodId, intent.umlType.id, true))
            };
        }));
}

/**
 * Creates a ViewStateReducer that changes the name of a new UmlClass and reduces its state
 * @param intent$ The original stream of intents
 */
export function cancelEditingUmlMethodReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof CancelEditingUmlMethodIntent)
        .map(intent => intent as CancelEditingUmlMethodIntent)
        .do(i => logReducer(log, "cancelEditingUmlMethodReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {
            return {
                menuItems: oldState.menuItems,
                umlDocument: oldState.umlDocument,
                currentInteractionTransaction: NoInteractionTransactionRunning.getInstance(),
                errors: oldState.errors,
                infoMessages: oldState.infoMessages,
                selectedElements: oldState.selectedElements.set(intent.methodId, new MethodSelectedElement(intent.methodId, intent.umlType.id, false))
            };
        }));
}


/**
 * Creates a ViewStateReducer that changes the name of a uml method and reduces the state
 * @param intent$ The original stream of intents
 */
export function changeUmlMethodSpecificationReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof ChangeUmlMethodIntent)
        .map(intent => intent as ChangeUmlMethodIntent)
        .do(i => logReducer(log, "changeUmlMethodSpecificationReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {
            const [document, errorMsg] = changeUmlMethodSpecification(oldState.umlDocument, intent.umlType.id, intent.methodId, intent.newMethodSpec)
            return {
                menuItems: oldState.menuItems,
                umlDocument: document,
                currentInteractionTransaction: NoInteractionTransactionRunning.getInstance(),
                errors: addOrRemoveError(oldState.errors, intent.methodId, errorMsg),
                infoMessages: oldState.infoMessages,
                selectedElements: oldState.selectedElements.set(intent.methodId, new MethodSelectedElement(intent.methodId, intent.umlType.id, false))

            };
        }));
}

/**
 * Creates a ViewStateReducer that changes the name of a uml method and reduces the state
 * @param intent$ The original stream of intents
 */
export function addUmlPropertyReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof AddUmlPropertyIntent)
        .map(intent => intent as AddUmlPropertyIntent)
        .do(i => logReducer(log, "addUmlPropertyReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {
            const propertyId = UUID.UUID()
            const [document, errorMsg] = addUmlProperty(oldState.umlDocument, intent.umlType.id, propertyId, "- property : int")
            return {
                menuItems: oldState.menuItems,
                umlDocument: document,
                currentInteractionTransaction: NoInteractionTransactionRunning.getInstance(),
                errors: addOrRemoveError(oldState.errors, propertyId, errorMsg),
                infoMessages: oldState.infoMessages,
                selectedElements: setMultipleSelected(
                    new UmlTypeSelected(intent.umlType.id, false),
                    new PropertySelectedElement(propertyId, intent.umlType.id, true))
            };
        }));
}


/**
 * Creates a ViewStateReducer that changes the name of a uml method and reduces the state
 * @param intent$ The original stream of intents
 */
export function startEditingUmlPropertyReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof StartEditingUmlPropertyIntent)
        .map(intent => intent as StartEditingUmlPropertyIntent)
        .do(i => logReducer(log, "startEditingUmlPropertyReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {
            return {
                menuItems: oldState.menuItems,
                umlDocument: oldState.umlDocument,
                currentInteractionTransaction: NoInteractionTransactionRunning.getInstance(),
                errors: oldState.errors,
                infoMessages: oldState.infoMessages,
                selectedElements: setMultipleSelected(
                    new UmlTypeSelected(intent.umlType.id, false),
                    new PropertySelectedElement(intent.propertyId, intent.umlType.id, true))
            };
        }));
}


/**
 * Creates a ViewStateReducer that changes the name of a uml method and reduces the state
 * @param intent$ The original stream of intents
 */
export function cancelEditingUmlPropertyReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof CancelEditingUmlPropertyIntent)
        .map(intent => intent as CancelEditingUmlPropertyIntent)
        .do(i => logReducer(log, "cancelEditingUmlPropertyReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {
            return {
                menuItems: oldState.menuItems,
                umlDocument: oldState.umlDocument,
                currentInteractionTransaction: NoInteractionTransactionRunning.getInstance(),
                errors: oldState.errors,
                infoMessages: oldState.infoMessages,
                selectedElements: oldState.selectedElements.set(intent.propertyId, new PropertySelectedElement(intent.propertyId, intent.umlType.id, false))
            };
        }));
}


/**
 * Creates a ViewStateReducer that changes the sepecification of a uml property and reduces the state
 * @param intent$ The original stream of intents
 */
export function changeUmlPropertyReducer(intent$: Rx.Observable<Intent>, log: boolean): Rx.Observable<ViewStateReducer> {
    return intent$
        .filter(intent => intent instanceof ChangeUmlPropertyIntent)
        .map(intent => intent as ChangeUmlPropertyIntent)
        .do(i => logReducer(log, "changeUmlPropertyReducer"))
        .switchMap(intent => Rx.Observable.of((oldState: ViewState): ViewState => {

            const [document, errorMsg] = changeUmlProperty(oldState.umlDocument, intent.umlType.id, intent.propertyId, intent.newPropertySpec)

            return {
                menuItems: oldState.menuItems,
                umlDocument: document,
                currentInteractionTransaction: NoInteractionTransactionRunning.getInstance(),
                errors: addOrRemoveError(oldState.errors, intent.propertyId, errorMsg),
                infoMessages: oldState.infoMessages,
                selectedElements: oldState.selectedElements.set(intent.propertyId, new PropertySelectedElement(intent.propertyId, intent.umlType.id, false))
            };
        }));
}