import * as React from "react";
import * as Rx from "@reactivex/rxjs";
import { EditorUmlDocument } from "../../businesslogic/model/EditorUmlDocument";
import { EditorUmlType } from "../../businesslogic/model/EditorUmlClass";
import { EditorUmlInheritance } from "../../businesslogic/model/EditorUmlInheritance";
import { ViewState, initialViewState } from "../../businesslogic/ViewState"
import { MenuButton } from "../menu/MenuButton"
import { MenuItem, MenuItemType } from "../../businesslogic/MenuItem"
import { Intent, IntentEmitter } from "../../businesslogic/intent/Intent"
import { FirstElementSelectedIntent, SecondElementSelectedIntent } from "../../businesslogic/intent/InteractionTransactionIntent"
import { ClickOnUmlCanvasIntent, DeleteSelectedElementsIntent, MoveSelectedElementsIntent, SelectUmlTypeIntent, CancelEverythingIntent } from "../../businesslogic/intent/UmlCanvasIntents"
import { UmlClass } from "../uml/UmlClass"
import { modelFunction } from "../../businesslogic/ModelFunc"
import { SelectingTwoElementsInteractionTransaction, MovingAnchorPointInteractionTransaction } from "../../businesslogic/interaction/InteractionTransaction"
import { LoadUmlDocumentIntent, SaveUmlDocumentIntent } from "../../businesslogic/intent/LoadSaveIntents"
import { StartAddingUmlTypeIntent, StartSelectingElementsIntent } from "../../businesslogic/intent/InteractionTransactionIntent"
import { UmlType } from "../../businesslogic/model/EditorUmlClass"
import { InfoMessageType, InfoMessage } from "../../businesslogic/InfoMessage"
import { UmlConnectionCanvas, UmlConnectionType } from "../uml/UmlConnectionCanvas"
import { isTypeSelected } from "../../businesslogic/SelectedElement"
import { isNoInteractionRunning } from "../../businesslogic/interaction/InteractionTransaction"

const KEY_BACKSPACE = 8
const KEY_DELETE = 46
const KEY_ESC = 27


export const CANVAS_WIDTH_PX = 3000
export const CANVAS_HEIGHT_PX = 3000

export class EditorProps {
    /**
     * The function to load a UmlDocument
     */
    loadModel: () => EditorUmlDocument

    /**
     * Save the model to 
     */
    saveModel: (EditorUmlDocument) => void

    // TODO error / test results
}

export class Editor extends React.Component<EditorProps, ViewState> {

    private readonly intentSubject = new Rx.Subject<Intent>();
    private readonly intentEmitter = new IntentEmitter(this.intentSubject);
    private subscription?: Rx.Subscription = null;
    private currentState?: ViewState = null; // For performance reasons
    private movingElement = false;
    private umlCanvasElement: HTMLElement = null;
    private distanceFromMouseToUmlClassLeftBorder = 0
    private distanceFromMouseToUmlClassTopBorder = 0


    constructor(props: EditorProps) {
        super(props);
        this.state = initialViewState();
        this.onClickInUmlCanvas = this.onClickInUmlCanvas.bind(this);
        this.onUmlTypeClicked = this.onUmlTypeClicked.bind(this);
        this.onMenuItemClicked = this.onMenuItemClicked.bind(this);
        this.handleKeyDown = this.handleKeyDown.bind(this);
        this.getBigInfoNotification = this.getBigInfoNotification.bind(this);
        this.isConnectionSelected = this.isConnectionSelected.bind(this);
        this.onMouseDownInUmlCanvas = this.onMouseDownInUmlCanvas.bind(this);
        this.onMouseUpInUmlCanvas = this.onMouseUpInUmlCanvas.bind(this);
        this.onMouseLeaveUmlCanvas = this.onMouseLeaveUmlCanvas.bind(this);
        this.onMouseMoveInUmlCanvas = this.onMouseMoveInUmlCanvas.bind(this);
        this.onMouseDownInUmlType = this.onMouseDownInUmlType.bind(this);
        this.getUmlCanvasElement = this.getUmlCanvasElement.bind(this);
    }

    render() {

        const menuItems = this.state.menuItems.map(
            item => <MenuButton key={item.type} menuItem={item} menuItemClicked={this.onMenuItemClicked} />
        );

        const allowAddingPropertiesAndMethods = (this.movingElement === false)
            && isNoInteractionRunning(this.state.currentInteractionTransaction)

        const umlClasses = this.state.umlDocument.types.map(
            umlClass => <UmlClass
                key={umlClass.id}
                intentEmitter={this.intentEmitter}
                umlType={umlClass}
                onClickListener={this.onUmlTypeClicked}
                onMouseDownListener={this.onMouseDownInUmlType}
                errors={this.state.errors}
                selectionMap={this.state.selectedElements}
                addMethodsAndPropertiesEnabled={allowAddingPropertiesAndMethods}
                interactionTransaction={this.state.currentInteractionTransaction}
            />
        ).toArray();

        const bigNotification = this.getBigInfoNotification()

        // TODO fix mouse event dispatching for siblings so that relations etc. can be drawn on top of uml classes
        // https://stackoverflow.com/questions/1009753/pass-mouse-events-through-absolutely-positioned-element
        // https://stackoverflow.com/questions/24082498/conditional-mouse-event-pass-through-between-sibling-elements-that-overlap
        return (
            <div className="editor-container" onKeyDown={this.handleKeyDown}>
                <div className="menu-container">
                    {menuItems}
                </div>
                <div id="uml-canvas" onClick={this.onClickInUmlCanvas} onMouseDown={this.onMouseDownInUmlCanvas} onMouseUp={this.onMouseUpInUmlCanvas} onMouseMove={this.onMouseMoveInUmlCanvas} onMouseLeave={this.onMouseLeaveUmlCanvas}>
                    <UmlConnectionCanvas
                        key="uml-connection-canvas"
                        selectedElementsMap={this.state.selectedElements}
                        umlDocument={this.state.umlDocument}
                        intentEmitter={this.intentEmitter}
                        errorsMap={this.state.errors}
                        movingAnchorPointInteractionTransaction={this.state.currentInteractionTransaction instanceof MovingAnchorPointInteractionTransaction ? this.state.currentInteractionTransaction : null}
                    />
                    {umlClasses}
                </div>
                {bigNotification !== null &&
                    <div key="BigNotifi0cation" className="toast-big">{bigNotification.message}</div>
                }
            </div>
        );
    }

    getBigInfoNotification(): InfoMessage {
        const lastAdded = this.state.infoMessages.filter(m => m.type == InfoMessageType.BIG_INFO).last()
        if (lastAdded == null) {
            return null
        } else {
            return lastAdded
        }
    }


    handleKeyDown(event: any) { // TODO what type is that event?
        const tag = event.target.tagName.toLowerCase();
        if (tag != 'input' && tag != 'textarea') {
            // Only intercept events if they are not input fileds
            if (event.keyCode == KEY_BACKSPACE || event.keyCode == KEY_DELETE) {
                this.intentEmitter.emit(new DeleteSelectedElementsIntent())
            } else if (event.keyCode == KEY_ESC) {
                this.intentEmitter.emit(new CancelEverythingIntent())
            }
        }
    }


    componentDidMount() {
        this.subscription = modelFunction(this.intentSubject as Rx.Observable<Intent>, this.props.saveModel, this.props.loadModel)
            .subscribe(
            (newState) => {
                // only renders if state has been changed
                if (this.currentState !== newState) {
                    this.setState(newState)
                    this.currentState = newState;
                }
            },
            (error) => {
                // TODO rerror handling
                // i.e. reload the page
            }
            );

        document.body.addEventListener("keydown", this.handleKeyDown);
    }


    componentWillUnmount() {
        if (this.subscription != null) {
            this.subscription.unsubscribe();
        }
        document.body.removeEventListener("keydown", this.handleKeyDown)
    }

    onMouseDownInUmlCanvas(event: any) {
    }

    onMouseUpInUmlCanvas(event: any) {
        this.movingElement = false
        this.forceUpdate()
    }

    onMouseMoveInUmlCanvas(event: any) {
        if (this.movingElement) {
            const umlCanvasElement = this.getUmlCanvasElement()
            var newX = event.pageX - umlCanvasElement.offsetLeft + umlCanvasElement.scrollLeft - this.distanceFromMouseToUmlClassLeftBorder;
            var newY = event.pageY - umlCanvasElement.offsetTop + umlCanvasElement.scrollTop - this.distanceFromMouseToUmlClassTopBorder;

            if (newX < 0) newX = 0
            if (newY < 0) newY = 0
            this.intentEmitter.emit(MoveSelectedElementsIntent.obtain(newX, newY))
        }
    }

    onMouseLeaveUmlCanvas(event: any) {
        this.forceUpdate()
    }

    getUmlCanvasElement(): HTMLElement {
        if (this.umlCanvasElement == null) {
            this.umlCanvasElement = document.getElementById("uml-canvas")
        }
        return this.umlCanvasElement
    }

    onMouseDownInUmlType(umlType: EditorUmlType, x: number, y: number, distanceFromMouseToUmlClassLeftBorder: number, distanceFromMouseToUmlClassTopBorder: number) {
        this.movingElement = true
        this.distanceFromMouseToUmlClassLeftBorder = distanceFromMouseToUmlClassLeftBorder
        this.distanceFromMouseToUmlClassTopBorder = distanceFromMouseToUmlClassTopBorder

        this.intentEmitter.emit(new SelectUmlTypeIntent(umlType));
    }



    onMenuItemClicked(menuType: MenuItemType) {

        switch (menuType) {
            case MenuItemType.CLASS:
                this.intentEmitter.emit(new StartAddingUmlTypeIntent(menuType, UmlType.CLASS))
                break;

            case MenuItemType.INTERFACE:
                this.intentEmitter.emit(new StartAddingUmlTypeIntent(menuType, UmlType.INTERFACE))
                break;

            case MenuItemType.INHERITANCE:
            case MenuItemType.AGGREGATION:
            case MenuItemType.ASSOCIATION:
            case MenuItemType.COMPOSITION:
            case MenuItemType.IMPLEMENT_INTERFACE:
                this.intentEmitter.emit(new StartSelectingElementsIntent(menuType))
                break;


            case MenuItemType.SAVE_AND_RUN:
                this.intentEmitter.emit(new SaveUmlDocumentIntent());
                break;

            // TODO implement others
            default: throw Error("Don't know what to do on click on MenuItem of type " + menuType);
        }
    }

    /**
     * Called when the user clicks somewhere in the canvas
     * @param event the clicke event
     */
    onClickInUmlCanvas(event: any) {
        const umlCanvasElement = this.getUmlCanvasElement()
        const x = event.pageX - umlCanvasElement.offsetLeft + umlCanvasElement.scrollLeft
        const y = event.pageY - umlCanvasElement.offsetTop + umlCanvasElement.scrollTop
        this.intentEmitter.emit(
            new ClickOnUmlCanvasIntent(this.state.currentInteractionTransaction, x, y));

        this.movingElement = false
    }

    onUmlTypeClicked(umlType: EditorUmlType) {
        const currentTransaction = this.state.currentInteractionTransaction;

        if (currentTransaction instanceof SelectingTwoElementsInteractionTransaction) {
            if (currentTransaction.firstSelectedElementId == null) {
                this.intentEmitter.emit(new FirstElementSelectedIntent(currentTransaction.menuItemType, umlType.id));
            } else {
                this.intentEmitter.emit(new SecondElementSelectedIntent(currentTransaction.menuItemType, currentTransaction.firstSelectedElementId, umlType.id));
            }
        } else {
            this.intentEmitter.emit(new SelectUmlTypeIntent(umlType));
        }


        // console.log("Clicked on "+umlType.name)
    }

    isConnectionSelected(id: string) {
        if (this.state.selectedElements.get(id) === undefined)
            return false
        else
            return true
    }
}