import * as React from "react";
import * as Rx from "@reactivex/rxjs";
import { EditorUmlType } from "../../businesslogic/model/EditorUmlClass";
import { IntentEmitter } from "../../businesslogic/intent/Intent";

const KEY_CODE_ENTER = 13;
const KEY_CODE_ESC = 27;

export type TextInputProps = {
    /**
     * The id to identify the callback. Typically this is the uml property id or the uml method id
     */
    readonly id: string,


    /**
     * The id type helps to identify in the callback whether it was a uml property or uml method which text has been changed
     */
    readonly idType: string,

    readonly text: string,
    readonly focused: boolean,
    readonly selected: boolean,

    /**
     * Should the text be displayed bold?
     */
    readonly bold: boolean,

    /**
     * Should an error be displayed? If null or undefined then no error is displayed
     */
    readonly errorMessage: string,

    /**
     * Define addirtional css styles that are applied to this one
     */
    readonly additionalStyleProps?: any

    /**
     * Called when the ENTER key has been pressed and the text has been changed compared to the original one
     */
    readonly onTextChanged: (id: String, idType: string, newText: String) => void,

    /**
     * Called when the ESC key has been pressed
     */
    readonly onChangingTextCanceled: (id: String, idType: string) => void

    /**
     * Called when the user wants to change this text (typically on double clicking the text element)
     */
    readonly onFocusRequired: (id: String, idType: string) => void

    /**
     * Called when the user clicks on the text (typically to select this item like a uml method definition)
     */
    readonly onClicked: (id: String, idType: string) => void

}

type TextInputState = {
    value: string
}

export class TextInput extends React.Component<TextInputProps, TextInputState> {

    // Workaround to request the focus
    // private inputElement: HTMLInputElement

    private enterPressed = false;

    constructor(props: TextInputProps) {
        super(props);
        this.state = { value: props.text }
        this.onKeyDown = this.onKeyDown.bind(this);
        this.onTextChanged = this.onTextChanged.bind(this);
        this.onDoubleClick = this.onDoubleClick.bind(this);
        this.onBlur = this.onBlur.bind(this);
        this.onClicked = this.onClicked.bind(this);
        this.onMouseDown = this.onMouseDown.bind(this);
    }

    render() {
        const computedWidth = ((this.state.value.length + 1) * 8)
        const computedWidthPx = computedWidth + "px";
        let styleProps = null;
        if (this.props.errorMessage != null && !this.props.focused) {
            styleProps = {
                width: computedWidthPx,
                color: "red"
            };
        } else {
            styleProps = {
                width: computedWidthPx
            };
        }

        if (this.props.bold === true) {
            styleProps.fontWeight = "bold"
        }

        // Copy additional styles if any specified
        var hardCodedWidth = -1
        if (this.props.additionalStyleProps != null) {
            for (var propName in this.props.additionalStyleProps) {
                styleProps[propName] = this.props.additionalStyleProps[propName]
                if (propName === "width") {
                    const widthDef: string = this.props.additionalStyleProps.width
                    hardCodedWidth = Number(widthDef.substring(0, widthDef.indexOf("px")))
                }
            }
        }

        // Override the computed witdh
        if (hardCodedWidth > computedWidth) {
            styleProps.width = this.props.additionalStyleProps.width
        }

        if (this.props.focused) {
            return (
                <input
                    autoFocus
                    type="text"
                    className="uml-input-field"
                    style={styleProps}
                    value={this.state.value}
                    onKeyDown={this.onKeyDown}
                    onChange={this.onTextChanged}
                    contentEditable={true}
                    disabled={false}
                    onBlur={this.onBlur}
                    onMouseDown={this.onMouseDown}
                    onClick={this.onClicked}
                />
            );

        } else {
            styleProps.userSelect = "none"
            if (this.props.selected === true) {
                styleProps.color = "blue"
            }
            return (
                <div
                    onDoubleClick={this.onDoubleClick}
                    onClick={this.onClicked}
                    title={this.props.errorMessage}
                >
                    <input
                        type="text"
                        // ref={(input) => { this.inputElement = input; }}
                        style={styleProps}
                        className="uml-input-field"
                        value={this.state.value}
                        onKeyDown={this.onKeyDown}
                        onChange={this.onTextChanged}
                        contentEditable={false}
                        disabled={true}
                        onBlur={this.onBlur}
                    />
                </div>
            );
        }
    }



    onKeyDown(event: any) { // TODO what type is that event?
        if (KEY_CODE_ENTER === event.keyCode) {
            if (event.target.value !== this.props.text) {
                // Text has been changed
                this.props.onTextChanged(this.props.id, this.props.idType, event.target.value);
            } else {
                // Text is the same as before - therefore we treat this a cancel event
                this.props.onChangingTextCanceled(this.props.id, this.props.idType);
            }
            event.preventDefault(); // prevent onTextChanged()
        } else if (KEY_CODE_ESC === event.keyCode) {
            this.props.onChangingTextCanceled(this.props.id, this.props.idType);
            this.setState({ value: this.props.text })
            event.preventDefault(); // prevent onTextChanged()
        }
        // Otherwise do nothing which triggers onTextChanged()
    }

    onMouseDown(event: any) {
        if (this.props.focused === true) {
            event.stopPropagation()
        }
    }


    onTextChanged(event: any) {
        this.setState({ value: event.target.value });
    }

    onDoubleClick(event: any) {
        this.props.onFocusRequired(this.props.id, this.props.idType);
    }

    onBlur(event: any) {
        if (event.target.value !== this.props.text) {
            this.props.onTextChanged(this.props.id, this.props.idType, event.target.value);
        }
    }

    onClicked(event: any) {
        if (this.props.focused === true) {
        } else {
            this.props.onClicked(this.props.id, this.props.idType)
        }

        event.stopPropagation()

    }

}

