import * as React from "react";
import * as Rx from "@reactivex/rxjs";
import { EditorUmlType, UmlType } from "../../businesslogic/model/EditorUmlClass";
import { TextInput } from "./TextInput";
import { IntentEmitter } from "../../businesslogic/intent/Intent";
import {
    RenameUmlTypeIntent, CancelRenameUmlTypeIntent, StartRenamingUmlTypeIntent, AddUmlMethodIntent,
    ChangeUmlMethodIntent, StartEditingUmlMethodIntent, CancelEditingUmlMethodIntent, AddUmlPropertyIntent,
    ChangeUmlPropertyIntent, StartEditingUmlPropertyIntent, CancelEditingUmlPropertyIntent
} from "../../businesslogic/intent/UmlTypeIntents";
import { SelectUmlTypeIntent, SelectUmlTypeAndMethodIntent, SelectUmlTypeAndPropertyIntent } from "../../businesslogic/intent/UmlCanvasIntents";
import { ErrorsMap } from "../../businesslogic/ErrorMessage"
import { SelectedElementsMap, UmlTypeSelected, PropertySelectedElement, MethodSelectedElement, isMethodSelected, isProperySelected, isTypeSelected } from "../../businesslogic/SelectedElement"
import { InteractionTransaction, NoInteractionTransactionRunning } from "../../businesslogic/interaction/InteractionTransaction"


const ID_TYPE_TITLE = "ClassNameTitle";
const ID_TYPE_PROPERTY = "Property";
const ID_TYPE_METHOD = "Method";

export type UmlClassProps = {
    readonly umlType: EditorUmlType,
    readonly intentEmitter: IntentEmitter,
    readonly onClickListener: (umlType: EditorUmlType) => void
    readonly onMouseDownListener: (umlType: EditorUmlType, x: number, y: number, distanceToLeftUmlClassBorder: number, distanceToTopUmlClassBorder) => void
    readonly errors: ErrorsMap,
    readonly selectionMap: SelectedElementsMap,
    readonly addMethodsAndPropertiesEnabled: boolean,
    readonly interactionTransaction: InteractionTransaction
}

type State = {
    showAddPropertiesButton: boolean,
    showAddMethodsButton: boolean
}
export class UmlClass extends React.Component<UmlClassProps, State> {

    private umlCanvasOffsetLeft: number
    private umlCanvasOffsetTop: number
    private propertyTextInputHasFocus = false;
    private methodTextInputHasFocus = false;

    constructor(props: UmlClassProps) {
        super(props);
        this.state = {
            showAddMethodsButton: false,
            showAddPropertiesButton: false
        };

        const umlCanvas = document.getElementById("uml-canvas");

        this.umlCanvasOffsetLeft = umlCanvas.offsetLeft;
        this.umlCanvasOffsetTop = umlCanvas.offsetTop;

        this.onTextCanceled = this.onTextCanceled.bind(this);
        this.onTextChanged = this.onTextChanged.bind(this);
        this.onTextInputClicked = this.onTextInputClicked.bind(this);
        this.onClicked = this.onClicked.bind(this);
        this.onMouseDown = this.onMouseDown.bind(this);
        this.onTextInputRequiresFocus = this.onTextInputRequiresFocus.bind(this);
        this.onMouseEnterPropertiesSection = this.onMouseEnterPropertiesSection.bind(this);
        this.onMouseLeavePropertySection = this.onMouseLeavePropertySection.bind(this);
        this.onMouseEnterMethodsSection = this.onMouseEnterMethodsSection.bind(this);
        this.onMouseLeaveMethodsSection = this.onMouseLeaveMethodsSection.bind(this);
        this.onAddMethodClicked = this.onAddMethodClicked.bind(this);
        this.onAddPropertyClicked = this.onAddPropertyClicked.bind(this);
        this.stopMouseEventPropagation = this.stopMouseEventPropagation.bind(this);
    }

    render() {
        const uml = this.props.umlType
        const isClass = this.props.umlType.type == UmlType.CLASS

        const errors = this.props.errors

        const position = {
            top: uml.y + "px",
            left: uml.x + "px"
        };

        let cssClasses = "uml-type-element"
        let sectionCssClasses = "uml-type-section"
        const [umlTypeSelected, umlTypeEditingName] = isTypeSelected(this.props.selectionMap, uml.id)
        if (!umlTypeSelected) {
            // unselected state
            cssClasses += (isClass ? " uml-type-class" : " uml-type-interface")
            sectionCssClasses += " uml-type-section"
        } else {
            // Selected state
            cssClasses += (isClass ? " uml-type-class-selected" : " uml-type-interface-selected")
            sectionCssClasses += " uml-type-section-selected"
        }

        let propertiesTextInputs = null
        this.propertyTextInputHasFocus = false;
        if (isClass) {
            propertiesTextInputs = this.props.umlType.properties.map(property => {

                const [propertySelected, propertyEditing] = isProperySelected(this.props.selectionMap, property.id)

                if (!this.propertyTextInputHasFocus && propertyEditing) {
                    this.propertyTextInputHasFocus = true;
                }

                return <TextInput
                    key={property.id}
                    id={property.id} idType={ID_TYPE_PROPERTY}
                    focused={propertyEditing}
                    selected={propertySelected}
                    text={property.propertySpec}
                    bold={false}
                    onClicked={this.onTextInputClicked}
                    onTextChanged={this.onTextChanged}
                    onChangingTextCanceled={this.onTextCanceled}
                    onFocusRequired={this.onTextInputRequiresFocus}
                    errorMessage={errors.get(property.id)} />
            }
            ).toArray();

            if (this.state.showAddPropertiesButton) {
                propertiesTextInputs.push(
                    <button key={"AddProperty" + this.props.umlType.id} className="uml-type-add-to-section-button" alt="add property" onClick={this.onAddPropertyClicked} onMouseDown={this.stopMouseEventPropagation} onMouseUp={this.stopMouseEventPropagation}>+</button>);
            }
        }

        this.methodTextInputHasFocus = false;
        const methodTextInputs = this.props.umlType.methods.map(method => {

            const [methodSelected, methodEditing] = isMethodSelected(this.props.selectionMap, method.id)

            if (!this.methodTextInputHasFocus && methodEditing) {
                this.methodTextInputHasFocus = true;
            }
            return (
                <TextInput
                    key={method.id}
                    id={method.id} idType={ID_TYPE_METHOD}
                    focused={methodEditing}
                    selected={methodSelected}
                    text={method.methodSepcification}
                    bold={false}
                    onClicked={this.onTextInputClicked}
                    onTextChanged={this.onTextChanged}
                    onChangingTextCanceled={this.onTextCanceled}
                    onFocusRequired={this.onTextInputRequiresFocus}
                    errorMessage={errors.get(method.id)}
                />)
        }
        ).toArray();



        if (this.state.showAddMethodsButton) {
            methodTextInputs.push(
                <button key={"AddMethod" + this.props.umlType.id} className="uml-type-add-to-section-button" alt="add method" onClick={this.onAddMethodClicked} onMouseDown={this.stopMouseEventPropagation} onMouseUp={this.stopMouseEventPropagation}>+</button>);
        }

        let selectedElement = this.props.selectionMap.get(uml.id)
        let editingTypeName = false
        if (selectedElement == null) {
            editingTypeName = false
        } else {
            const typeSelected = selectedElement as UmlTypeSelected
            editingTypeName = typeSelected.editingName
        }

        const sectionEnabled = this.props.addMethodsAndPropertiesEnabled
        const onMouseEnterPropertiesSection = sectionEnabled === true ? this.onMouseEnterPropertiesSection : null
        const onMouseLeavePropertySection = sectionEnabled === true ? this.onMouseLeavePropertySection : null
        const onMouseEnterMethodsSection = sectionEnabled === true ? this.onMouseEnterMethodsSection : null
        const onMouseLeaveMethodsSection = sectionEnabled === true ? this.onMouseLeaveMethodsSection : null

        return (
            <div id={this.props.umlType.id} className={cssClasses} style={position}
                onClick={this.onClicked} onMouseDown={this.onMouseDown}>
                <div className="uml-type-header">
                    <div className="uml-type-name">
                        <TextInput
                            key={"title" + this.props.umlType.id}
                            id={"title"} idType={ID_TYPE_TITLE}
                            focused={umlTypeEditingName}
                            selected={false}
                            text={this.props.umlType.name}
                            bold={true}
                            onClicked={this.onTextInputClicked}
                            onTextChanged={this.onTextChanged}
                            onChangingTextCanceled={this.onTextCanceled}
                            onFocusRequired={this.onTextInputRequiresFocus}
                            errorMessage={errors.get(this.props.umlType.id)} />
                    </div>
                </div>
                {!isClass ? null :
                    <div className={sectionCssClasses} onMouseEnter={onMouseEnterPropertiesSection} onMouseLeave={onMouseLeavePropertySection}>
                        {propertiesTextInputs}
                    </div>
                }
                <div className={sectionCssClasses} onMouseEnter={onMouseEnterMethodsSection} onMouseLeave={onMouseLeaveMethodsSection}>
                    {methodTextInputs}
                </div>
            </div>
        );
    }

    onTextChanged(id: string, idType: string, newText: string) {

        const intentEmitter = this.props.intentEmitter;
        if (idType === ID_TYPE_TITLE) {
            intentEmitter.emit(new RenameUmlTypeIntent(this.props.umlType, newText))
        }
        else if (idType === ID_TYPE_METHOD) {
            intentEmitter.emit(new ChangeUmlMethodIntent(this.props.umlType, id, newText))
        }
        else if (idType === ID_TYPE_PROPERTY) {
            intentEmitter.emit(new ChangeUmlPropertyIntent(this.props.umlType, id, newText))
        }
        else throw new Error("onTextInputRequiresFocus() on unknown idType " + idType)

    }

    onTextCanceled(id: string, idType: string) {
        const intentEmitter = this.props.intentEmitter;
        if (idType === ID_TYPE_TITLE) {
            intentEmitter.emit(new CancelRenameUmlTypeIntent(this.props.umlType))
        }
        else if (idType === ID_TYPE_METHOD) {
            intentEmitter.emit(new CancelEditingUmlMethodIntent(this.props.umlType, id))
        }
        else if (idType === ID_TYPE_PROPERTY) {
            intentEmitter.emit(new CancelEditingUmlPropertyIntent(this.props.umlType, id))
        }
        else throw new Error("onTextInputRequiresFocus() on unknown idType " + idType)
    }

    onClicked(event: any) {
        this.props.onClickListener(this.props.umlType);
        event.stopPropagation();
    }

    onMouseDown(event: any) {
        const diffToLeftBorder = event.pageX - this.umlCanvasOffsetLeft - this.props.umlType.x
        const diffToTopBorder = event.pageY - this.umlCanvasOffsetTop - this.props.umlType.y
        this.props.onMouseDownListener(this.props.umlType, event.pageX, event.pageY, diffToLeftBorder, diffToTopBorder)
    }


    onTextInputRequiresFocus(id: string, idType: string) {
        const intentEmitter = this.props.intentEmitter;
        if (idType === ID_TYPE_TITLE) {
            intentEmitter.emit(new StartRenamingUmlTypeIntent(this.props.umlType));
        }
        else if (idType === ID_TYPE_METHOD) {
            intentEmitter.emit(new StartEditingUmlMethodIntent(this.props.umlType, id))
        }
        else if (idType === ID_TYPE_PROPERTY) {
            intentEmitter.emit(new StartEditingUmlPropertyIntent(this.props.umlType, id))
        }
        else throw new Error("onTextInputRequiresFocus() on unknown idType " + idType)
    }

    onTextInputClicked(id: string, idType: string) {

        const intentEmitter = this.props.intentEmitter;
        const [umlTypeSelected, umlTypeEditingName] = isTypeSelected(this.props.selectionMap, this.props.umlType.id)

        if (this.props.interactionTransaction === NoInteractionTransactionRunning.getInstance()) {

            if (!umlTypeSelected || idType === ID_TYPE_TITLE) {
                intentEmitter.emit(new SelectUmlTypeIntent(this.props.umlType));
            }
            else if (idType === ID_TYPE_METHOD) {
                intentEmitter.emit(new SelectUmlTypeAndMethodIntent(this.props.umlType, id))
            }
            else if (idType === ID_TYPE_PROPERTY) {
                intentEmitter.emit(new SelectUmlTypeAndPropertyIntent(this.props.umlType, id))
            }
            else throw new Error("onTextInputRequiresFocus() on unknown idType " + idType)
        } else {
            this.props.onClickListener(this.props.umlType)
        }

    }

    onMouseEnterPropertiesSection(event: any) {
        const addPropertyButtonVisible = this.state.showAddPropertiesButton;
        if (!addPropertyButtonVisible && !this.propertyTextInputHasFocus) {
            const oldState = this.state;
            this.setState({
                showAddMethodsButton: false,
                showAddPropertiesButton: true
            });
        }
    }

    onMouseLeavePropertySection(event: any) {
        const addPropertyButtonVisible = this.state.showAddPropertiesButton;
        if (addPropertyButtonVisible) {
            const oldState = this.state;
            this.setState({
                showAddMethodsButton: false,
                showAddPropertiesButton: false
            });
        }
    }

    onMouseEnterMethodsSection(event: any) {
        const addMethodButtonVisible = this.state.showAddMethodsButton;
        if (!addMethodButtonVisible && !this.methodTextInputHasFocus) {
            const oldState = this.state;
            this.setState({
                showAddMethodsButton: true,
                showAddPropertiesButton: false
            });
        }

    }

    onMouseLeaveMethodsSection(event: any) {
        const addMethodsButtonVisible = this.state.showAddMethodsButton;
        if (addMethodsButtonVisible) {
            const oldState = this.state;
            this.setState({
                showAddMethodsButton: false,
                showAddPropertiesButton: false
            });
        }
    }

    stopMouseEventPropagation(event: any) {
        event.stopPropagation()
    }

    onAddMethodClicked(event: any) {
        this.props.intentEmitter.emit(new AddUmlMethodIntent(this.props.umlType))
        event.stopPropagation();
    }


    onAddPropertyClicked(event: any) {
        this.props.intentEmitter.emit(new AddUmlPropertyIntent(this.props.umlType))
        event.stopPropagation();
    }


    componentWillReceiveProps(nextProps: UmlClassProps) {
        if (nextProps.addMethodsAndPropertiesEnabled === false && (
            this.state.showAddMethodsButton === true || this.state.showAddPropertiesButton === true)) {
            this.setState({
                showAddMethodsButton: false,
                showAddPropertiesButton: false
            });
        }
    }


}

