import * as React from "react";
import * as Rx from "@reactivex/rxjs";
import { AnchorPoint, AnchorPointSide } from "../../businesslogic/model/AnchorPoint";
import { IntentEmitter } from "../../businesslogic/intent/Intent";
import { Layer, Rect, Circle, Stage, Group, Line } from 'react-konva';
import { Intent } from "../../businesslogic/intent/Intent"
import { SelectedElementsMap, SelectUmlRelation } from "../../businesslogic/SelectedElement"
import { ErrorMessage, ErrorsMap } from "../../businesslogic/ErrorMessage"
import {
    SelectUmlInheritanceIntent, SelectUmlInterfaceImplementationIntent, SelectUmlRelationIntent,
    DoubleClickUmlConnectionIntent, CancelEditingUmlRelationTextFieldIntent, ChangeUmlRelationTextFieldIntent,
    StartEditingUmlRelationTextFieldIntent, UmlRelationTextFieldType,
    StartMovingAnchorPointIntent, StopMovingAnchorPointIntent, MoveAnchorPointIntent
} from "../../businesslogic/intent/UmlCanvasIntents"
import { EditorUmlInheritance } from "../../businesslogic/model/EditorUmlInheritance"
import { Point } from "../../businesslogic/model/Point"
import { EditorUmlRelation, UmlRelationType } from "../../businesslogic/model/EditorUmlRelation"
import { EditorUmlInterfaceImplementation } from "../../businesslogic/model/EditorUmlInterfaceImplementation"
import { TextInput } from "./TextInput"
import { OrderedSet } from "immutable"
import { umlRelationMultiplicityId, EditorUmlDocument } from "../../businesslogic/model/EditorUmlDocument"
import { MovingAnchorPointInteractionTransaction } from "../../businesslogic/interaction/InteractionTransaction"
import { CANVAS_WIDTH_PX, CANVAS_HEIGHT_PX } from "../editor/Editor"



type Positions = {
    firstElementX: number,
    firstElementY: number
    firstElementWidth: number,
    firstElementHeight: number,
    secondElementX: number,
    secondElementY: number,
    secondElementWidth: number,
    secondElementHeight: number,
    firstElementAnchorPointXInCanvas: number,
    firstElementAnchorPointYInCanvas: number,
    firstAnchorPointSide: AnchorPointSide,
    secondElementAnchorPointXInCanvas: number,
    secondElementAnchorPointYInCanvas: number,
    secondAnchorPointSide: AnchorPointSide,
    selfReferenceAdditionalPoints?: Array<number>
}

type Line = {
    x1: number,
    y1: number,

    x2: number,
    y2: number
}

export type UmlConnection = EditorUmlInheritance | EditorUmlInterfaceImplementation | EditorUmlRelation


export type UmlConnectionProps = {

    umlDocument: EditorUmlDocument,

    /**
     * is the connection selected
     */
    selectedElementsMap: SelectedElementsMap,

    /**
     * Lookup map to display errors
     */
    errorsMap: ErrorsMap,
    /**
     * Emitter to fire new intents
     */
    intentEmitter: IntentEmitter,

    /**
     * This transaction indicates whether or not we are currently moving an anchor point
     */
    movingAnchorPointInteractionTransaction?: MovingAnchorPointInteractionTransaction
}


/**
 *  Icon size (in px) like triangle for inheritance
 */
const ICON_SIZE = 20

/**
 * The stroke width of the line (in px)
 */
const LINE_STROKE_WIDTH = 2

/**
 * Add so many pixels to handle touch events for lines
 */
const LINE_TOUCH_AREA_INCREASEMENT = 25


const FIRST_MULTIPLICITY_INPUT = "FirstMulitplicityInputField"
const SECOND_MULTIPLICITY_INPUT = "SecondMulitplicityInputField"

const transparentLineColor = "transparent"

const CHAR_WIDTH = 10


const ANCHOR_POINT_LOOKUP: Array<[AnchorPointSide, AnchorPointSide]> = [
    [AnchorPointSide.TOP, AnchorPointSide.TOP],
    [AnchorPointSide.TOP, AnchorPointSide.LEFT],
    [AnchorPointSide.TOP, AnchorPointSide.BOTTOM],
    [AnchorPointSide.TOP, AnchorPointSide.RIGHT],

    [AnchorPointSide.LEFT, AnchorPointSide.TOP],
    [AnchorPointSide.LEFT, AnchorPointSide.LEFT],
    [AnchorPointSide.LEFT, AnchorPointSide.BOTTOM],
    [AnchorPointSide.LEFT, AnchorPointSide.RIGHT],

    [AnchorPointSide.BOTTOM, AnchorPointSide.TOP],
    [AnchorPointSide.BOTTOM, AnchorPointSide.LEFT],
    [AnchorPointSide.BOTTOM, AnchorPointSide.BOTTOM],
    [AnchorPointSide.BOTTOM, AnchorPointSide.RIGHT],

    [AnchorPointSide.RIGHT, AnchorPointSide.TOP],
    [AnchorPointSide.RIGHT, AnchorPointSide.LEFT],
    [AnchorPointSide.RIGHT, AnchorPointSide.BOTTOM],
    [AnchorPointSide.RIGHT, AnchorPointSide.RIGHT]
]



export enum UmlConnectionType {
    INHERITANCE,
    INTERFACE_IMPLEMENTATION,
    RELATION // ASSOCIATION, AGGREGATION, COMPOSITION
}



/**
 * This elements adds a canvas to draw connections between types.
 * Since each connection could have a icon which could be displayed on every side,
 * we add an additional ICON_SIZE on each side of the canvas
 */
export class UmlConnectionCanvas extends React.Component<UmlConnectionProps, undefined> {

    private intentToDeliver?: Intent = null
    private renderSubscription?: Rx.Subscription = null
    private parentContainer?: HTMLElement = null
    private mouseUpDeliveredBeforeCatchingClick = false

    constructor(props: UmlConnectionProps) {
        super(props);
        this.intentToDeliver = null
        this.onClickedOnObjectOnTheCanvas = this.onClickedOnObjectOnTheCanvas.bind(this)
        this.onClickedOnContainer = this.onClickedOnContainer.bind(this)
        this.onDoubleClickedOnContainer = this.onDoubleClickedOnContainer.bind(this)
        this.onDoubleClickedOnObjectOnTheCanvas = this.onDoubleClickedOnObjectOnTheCanvas.bind(this)
        this.renderCanvas = this.renderCanvas.bind(this)
        this.areDependingElementsReady = this.areDependingElementsReady.bind(this)
        this.runRenderingOnceDependingUmlElementsAreReady = this.runRenderingOnceDependingUmlElementsAreReady.bind(this)
        this.onTextInputClicked = this.onTextInputClicked.bind(this)
        this.onTextInputRequiresFocus = this.onTextInputRequiresFocus.bind(this)
        this.onTextInputCanceled = this.onTextInputCanceled.bind(this)
        this.onTextInputChanged = this.onTextInputChanged.bind(this)
        this.renderInheritanceOrInterfaceImplementation = this.renderInheritanceOrInterfaceImplementation.bind(this)
        this.renderAssociation = this.renderAssociation.bind(this)
        this.renderCompositionOrAggregation = this.renderCompositionOrAggregation.bind(this)
        this.renderMultiplicity = this.renderMultiplicity.bind(this)
        this.getElement = this.getElement.bind(this)
        this.onMouseDownOnAnchorPoint = this.onMouseDownOnAnchorPoint.bind(this)
        this.onMouseMoveInCanvas = this.onMouseMoveInCanvas.bind(this)
        this.onMouseDownInCanvas = this.onMouseDownInCanvas.bind(this)
        this.onMouseUpInCanvas = this.onMouseUpInCanvas.bind(this)
        this.getParentContainer = this.getParentContainer.bind(this)
    }

    componentWillUnmount() {
        if (this.renderSubscription !== null) {
            this.renderSubscription.unsubscribe()
        }
    }


    render() {
        return this.runRenderingOnceDependingUmlElementsAreReady()
    }


    getElement(domId: string): HTMLElement {
        return document.getElementById(domId)
    }

    renderCanvas() {
        const cssClass = "uml-connection-canvas-container"

        const shapes: Array<JSX.Element> = []
        var multiplicityElements: Array<JSX.Element> = []

        //
        // Relations
        //
        this.props.umlDocument.relations.forEach(rel => {

            const selectedElement = this.props.selectedElementsMap.get(rel.id)
            const selected = selectedElement == null ? false : true
            var editingStartMultiplicity = false
            var editingEndMultiplicity = false
            var lineColor = "black"
            if (selected === true) {
                editingStartMultiplicity = (selectedElement as SelectUmlRelation).editingStartMultiplicity
                editingEndMultiplicity = (selectedElement as SelectUmlRelation).editingEndMultiplicity
                lineColor = "blue"
            }

            const isCompositionOrAggregation = rel.relationType === UmlRelationType.COMPOSITION || rel.relationType === UmlRelationType.AGGREGATION
            const positions = calculatePosition(this.getElement(rel.startTypeId), this.getElement(rel.endTypeId), rel.startTypeAnchorPoint, rel.endTypeAnchorPoint);


            if (isCompositionOrAggregation) {
                shapes.push(this.renderCompositionOrAggregation(rel, positions, lineColor, selected))
            } else {
                shapes.push(this.renderAssociation(rel, positions, lineColor, selected))
            }

            multiplicityElements = multiplicityElements.concat(this.renderMultiplicity(rel, positions, selected, editingStartMultiplicity, editingEndMultiplicity))
        });


        //
        // Inheritance
        // 
        this.props.umlDocument.inheritance.forEach(inh => {
            const positions = calculatePosition(this.getElement(inh.subTypeId), this.getElement(inh.superTypeId), inh.subTypeAnchorPoint, inh.superTypeAnchorPoint);

            const selectedElement = this.props.selectedElementsMap.get(inh.id)
            const selected = selectedElement == null ? false : true
            var lineColor = (selected === true) ? "blue" : " black"
            shapes.push(this.renderInheritanceOrInterfaceImplementation(inh, UmlConnectionType.INHERITANCE, positions, lineColor, selected))
        });

        //
        // Interface implementation
        //
        this.props.umlDocument.classImplementingInterface.forEach(impl => {
            const positions = calculatePosition(this.getElement(impl.classId), this.getElement(impl.interfaceId), impl.classAnchorPoint, impl.interfaceAnchorPoint);

            const selectedElement = this.props.selectedElementsMap.get(impl.id)
            const selected = selectedElement == null ? false : true
            var lineColor = (selected === true) ? "blue" : " black"
            shapes.push(this.renderInheritanceOrInterfaceImplementation(impl, UmlConnectionType.INTERFACE_IMPLEMENTATION, positions, lineColor, selected))
        });

        return (
            <div className={cssClass} onClick={this.onClickedOnContainer} onDoubleClick={this.onDoubleClickedOnContainer} onMouseMove={this.onMouseMoveInCanvas} onMouseDown={this.onMouseDownInCanvas} onMouseUp={this.onMouseUpInCanvas}>
                <Stage key="uml-canvas-stage" width={CANVAS_WIDTH_PX} height={CANVAS_HEIGHT_PX}>
                    <Layer key="uml-canvas-layer">
                        {shapes}
                    </Layer>
                </Stage>
                {multiplicityElements}
            </div >
        )
    }


    renderMultiplicity(connection: EditorUmlRelation, position: Positions, selected: boolean, editingStartMultiplicity: boolean, editingEndMultiplicity: boolean): Array<JSX.Element> {
        var firstMultiplicityPos = null;
        var secondMultiplicityPos = null;

        switch (position.firstAnchorPointSide) {
            case AnchorPointSide.LEFT:
                const textWidth = 6 * CHAR_WIDTH
                firstMultiplicityPos = {
                    left: position.firstElementAnchorPointXInCanvas - textWidth,
                    top: position.firstElementAnchorPointYInCanvas - ICON_SIZE - 5,
                    textAlign: "right",
                    width: textWidth + "px"
                }
                break;

            case AnchorPointSide.RIGHT:
                firstMultiplicityPos = {
                    left: position.firstElementAnchorPointXInCanvas + 5,
                    top: position.firstElementAnchorPointYInCanvas - ICON_SIZE - 5
                }
                break;


            case AnchorPointSide.TOP:
                firstMultiplicityPos = {
                    left: position.firstElementAnchorPointXInCanvas + (ICON_SIZE / 2),
                    top: position.firstElementAnchorPointYInCanvas - 20
                }
                break;

            case AnchorPointSide.BOTTOM:
                firstMultiplicityPos = {
                    left: position.firstElementAnchorPointXInCanvas + (ICON_SIZE / 2),
                    top: position.firstElementAnchorPointYInCanvas + 5
                }
                break;
        }



        switch (position.secondAnchorPointSide) {
            case AnchorPointSide.LEFT:
                const textWidth = 6 * CHAR_WIDTH
                secondMultiplicityPos = {
                    left: position.secondElementAnchorPointXInCanvas - textWidth,
                    top: position.secondElementAnchorPointYInCanvas - ICON_SIZE - 5,
                    textAlign: "right",
                    width: textWidth + "px"
                }
                break;

            case AnchorPointSide.RIGHT:
                secondMultiplicityPos = {
                    left: position.secondElementAnchorPointXInCanvas + 5,
                    top: position.secondElementAnchorPointYInCanvas - ICON_SIZE - 5
                }
                break;


            case AnchorPointSide.TOP:
                secondMultiplicityPos = {
                    left: position.secondElementAnchorPointXInCanvas + ICON_SIZE / 2,
                    top: position.secondElementAnchorPointYInCanvas - 20
                }
                break;

            case AnchorPointSide.BOTTOM:
                secondMultiplicityPos = {
                    left: position.secondElementAnchorPointXInCanvas + ICON_SIZE / 2,
                    top: position.secondElementAnchorPointYInCanvas + 5
                }
                break;
        }

        firstMultiplicityPos.position = "absolute"
        secondMultiplicityPos.position = "absolute"

        const startMultiplicityErrorMsg = this.props.errorsMap.get(umlRelationMultiplicityId(connection.id, true), null)
        const endMultiplicityErrorMsg = this.props.errorsMap.get(umlRelationMultiplicityId(connection.id, false), null)

        const elements: JSX.Element[] = [
            (<TextInput
                key={connection.id + "start-multiplicity"}
                idType={FIRST_MULTIPLICITY_INPUT}
                id={connection.id}
                errorMessage={startMultiplicityErrorMsg}
                onClicked={this.onTextInputClicked}
                onChangingTextCanceled={this.onTextInputCanceled}
                onFocusRequired={this.onTextInputRequiresFocus}
                onTextChanged={this.onTextInputChanged}
                bold={false}
                additionalStyleProps={firstMultiplicityPos}
                text={connection.startMultiplicity}
                focused={editingStartMultiplicity}
                selected={selected}
            />
            ),
            (<TextInput
                key={connection.id + "end-multiplicity"}
                idType={SECOND_MULTIPLICITY_INPUT}
                id={connection.id}
                errorMessage={endMultiplicityErrorMsg}
                onClicked={this.onTextInputClicked}
                onChangingTextCanceled={this.onTextInputCanceled}
                onFocusRequired={this.onTextInputRequiresFocus}
                onTextChanged={this.onTextInputChanged}
                bold={false}
                additionalStyleProps={secondMultiplicityPos}
                text={connection.endMultiplicity}
                focused={editingEndMultiplicity}
                selected={selected}
            />
            )
        ]

        return elements;

    }

    renderInheritanceOrInterfaceImplementation(connection: UmlConnection, umlConnectionType: UmlConnectionType, positions: Positions, lineColor: string, selected: boolean): JSX.Element {

        var triangle: Array<number> = null
        var line: Array<number> = null
        const dash: Array<number> = (umlConnectionType === UmlConnectionType.INHERITANCE) ? [] : [6, 3]

        switch (positions.secondAnchorPointSide) {
            case AnchorPointSide.LEFT:
                triangle = [
                    positions.secondElementAnchorPointXInCanvas,
                    positions.secondElementAnchorPointYInCanvas,

                    positions.secondElementAnchorPointXInCanvas - ICON_SIZE,
                    positions.secondElementAnchorPointYInCanvas - (ICON_SIZE / 2),

                    positions.secondElementAnchorPointXInCanvas - ICON_SIZE,
                    positions.secondElementAnchorPointYInCanvas + (ICON_SIZE / 2)
                ]
                line = [
                    positions.firstElementAnchorPointXInCanvas,
                    positions.firstElementAnchorPointYInCanvas,

                    positions.secondElementAnchorPointXInCanvas - ICON_SIZE,
                    positions.secondElementAnchorPointYInCanvas
                ]
                break;

            case AnchorPointSide.RIGHT:
                triangle = [
                    positions.secondElementAnchorPointXInCanvas,
                    positions.secondElementAnchorPointYInCanvas,

                    positions.secondElementAnchorPointXInCanvas + ICON_SIZE,
                    positions.secondElementAnchorPointYInCanvas - (ICON_SIZE / 2),

                    positions.secondElementAnchorPointXInCanvas + ICON_SIZE,
                    positions.secondElementAnchorPointYInCanvas + (ICON_SIZE / 2)
                ]
                line = [
                    positions.firstElementAnchorPointXInCanvas,
                    positions.firstElementAnchorPointYInCanvas,

                    positions.secondElementAnchorPointXInCanvas + ICON_SIZE,
                    positions.secondElementAnchorPointYInCanvas
                ]
                break;


            case AnchorPointSide.TOP:
                triangle = [
                    positions.secondElementAnchorPointXInCanvas,
                    positions.secondElementAnchorPointYInCanvas,

                    positions.secondElementAnchorPointXInCanvas - (ICON_SIZE / 2),
                    positions.secondElementAnchorPointYInCanvas - ICON_SIZE,

                    positions.secondElementAnchorPointXInCanvas + (ICON_SIZE / 2),
                    positions.secondElementAnchorPointYInCanvas - ICON_SIZE
                ]
                line = [
                    positions.firstElementAnchorPointXInCanvas,
                    positions.firstElementAnchorPointYInCanvas,

                    positions.secondElementAnchorPointXInCanvas,
                    positions.secondElementAnchorPointYInCanvas - ICON_SIZE
                ]
                break;

            case AnchorPointSide.BOTTOM:
                triangle = [
                    positions.secondElementAnchorPointXInCanvas,
                    positions.secondElementAnchorPointYInCanvas,

                    positions.secondElementAnchorPointXInCanvas - (ICON_SIZE / 2),
                    positions.secondElementAnchorPointYInCanvas + ICON_SIZE,

                    positions.secondElementAnchorPointXInCanvas + (ICON_SIZE / 2),
                    positions.secondElementAnchorPointYInCanvas + ICON_SIZE
                ]
                line = [
                    positions.firstElementAnchorPointXInCanvas,
                    positions.firstElementAnchorPointYInCanvas,

                    positions.secondElementAnchorPointXInCanvas,
                    positions.secondElementAnchorPointYInCanvas + ICON_SIZE
                ]
                break;
        }

        const self = this
        const clickFunc = event => self.onClickedOnObjectOnTheCanvas(event, connection, umlConnectionType)
        var mouseDownOnFirstAnchorPointFunc = null
        var mouseDownOnSecondAnchorPointFunc = null
        if (selected) {
            mouseDownOnFirstAnchorPointFunc = event => self.onMouseDownOnAnchorPoint(event, connection.id, true, positions)
            mouseDownOnSecondAnchorPointFunc = event => self.onMouseDownOnAnchorPoint(event, connection.id, false, positions)
        }

        return (
            <Group key={connection.id}>
                <Line // workaround to increase the touchable area of a line
                    points={line}
                    strokeWidth={LINE_TOUCH_AREA_INCREASEMENT}
                    stroke={transparentLineColor}
                    onClick={clickFunc}
                />
                <Line
                    points={line}
                    strokeWidth={LINE_STROKE_WIDTH}
                    stroke={lineColor}
                    dash={dash}
                    onClick={clickFunc}
                />
                <Line
                    points={triangle}
                    closed={true}
                    fill={"white"}
                    strokeWidth={LINE_STROKE_WIDTH}
                    stroke={lineColor}
                    onClick={clickFunc}
                />

                {selected &&
                    <Circle
                        x={positions.firstElementAnchorPointXInCanvas}
                        y={positions.firstElementAnchorPointYInCanvas}
                        radius={4}
                        stroke={"black"}
                        fill={"white"}
                        strokeWidth={4}
                        onMouseDown={mouseDownOnFirstAnchorPointFunc}
                    />
                }
                {selected &&
                    <Circle
                        x={positions.secondElementAnchorPointXInCanvas}
                        y={positions.secondElementAnchorPointYInCanvas}
                        radius={4}
                        stroke={"black"}
                        fill={"white"}
                        strokeWidth={4}
                        onMouseDown={mouseDownOnSecondAnchorPointFunc}

                    />
                }

            </Group>
        )
    }

    renderCompositionOrAggregation(connection: UmlConnection, positions: Positions, lineColor: string, selected: boolean): JSX.Element {

        var icon: Array<number> = null
        var line: Array<number> = null
        const iconFill = (connection as EditorUmlRelation).relationType === UmlRelationType.COMPOSITION ? "black" : "white"

        switch (positions.secondAnchorPointSide) {
            case AnchorPointSide.LEFT:
                icon = [
                    positions.secondElementAnchorPointXInCanvas,
                    positions.secondElementAnchorPointYInCanvas,

                    positions.secondElementAnchorPointXInCanvas - (ICON_SIZE / 2),
                    positions.secondElementAnchorPointYInCanvas - (ICON_SIZE / 4),

                    positions.secondElementAnchorPointXInCanvas - ICON_SIZE,
                    positions.secondElementAnchorPointYInCanvas,

                    positions.secondElementAnchorPointXInCanvas - (ICON_SIZE / 2),
                    positions.secondElementAnchorPointYInCanvas + (ICON_SIZE / 4)
                ]
                line = [
                    positions.firstElementAnchorPointXInCanvas,
                    positions.firstElementAnchorPointYInCanvas]

                if (positions.selfReferenceAdditionalPoints != null) {
                    line = line.concat(positions.selfReferenceAdditionalPoints)
                }


                line = line.concat([positions.secondElementAnchorPointXInCanvas - ICON_SIZE,
                positions.secondElementAnchorPointYInCanvas
                ])
                break;

            case AnchorPointSide.RIGHT:
                icon = [
                    positions.secondElementAnchorPointXInCanvas,
                    positions.secondElementAnchorPointYInCanvas,

                    positions.secondElementAnchorPointXInCanvas + (ICON_SIZE / 2),
                    positions.secondElementAnchorPointYInCanvas - (ICON_SIZE / 4),

                    positions.secondElementAnchorPointXInCanvas + ICON_SIZE,
                    positions.secondElementAnchorPointYInCanvas,

                    positions.secondElementAnchorPointXInCanvas + (ICON_SIZE / 2),
                    positions.secondElementAnchorPointYInCanvas + (ICON_SIZE / 4)
                ]
                line = [
                    positions.firstElementAnchorPointXInCanvas,
                    positions.firstElementAnchorPointYInCanvas
                ]

                if (positions.selfReferenceAdditionalPoints != null) {
                    line = line.concat(positions.selfReferenceAdditionalPoints)
                }

                line = line.concat([
                    positions.secondElementAnchorPointXInCanvas + ICON_SIZE,
                    positions.secondElementAnchorPointYInCanvas
                ])
                break;


            case AnchorPointSide.TOP:
                icon = [
                    positions.secondElementAnchorPointXInCanvas,
                    positions.secondElementAnchorPointYInCanvas,

                    positions.secondElementAnchorPointXInCanvas - (ICON_SIZE / 4),
                    positions.secondElementAnchorPointYInCanvas - (ICON_SIZE / 2),

                    positions.secondElementAnchorPointXInCanvas,
                    positions.secondElementAnchorPointYInCanvas - ICON_SIZE,

                    positions.secondElementAnchorPointXInCanvas + (ICON_SIZE / 4),
                    positions.secondElementAnchorPointYInCanvas - (ICON_SIZE / 2)
                ]
                line = [
                    positions.firstElementAnchorPointXInCanvas,
                    positions.firstElementAnchorPointYInCanvas
                ]

                if (positions.selfReferenceAdditionalPoints != null) {
                    line = line.concat(positions.selfReferenceAdditionalPoints)
                }

                line = line.concat([
                    positions.secondElementAnchorPointXInCanvas,
                    positions.secondElementAnchorPointYInCanvas - ICON_SIZE
                ])
                break;

            case AnchorPointSide.BOTTOM:
                icon = [
                    positions.secondElementAnchorPointXInCanvas,
                    positions.secondElementAnchorPointYInCanvas,

                    positions.secondElementAnchorPointXInCanvas - (ICON_SIZE / 4),
                    positions.secondElementAnchorPointYInCanvas + (ICON_SIZE / 2),

                    positions.secondElementAnchorPointXInCanvas,
                    positions.secondElementAnchorPointYInCanvas + ICON_SIZE,

                    positions.secondElementAnchorPointXInCanvas + (ICON_SIZE / 4),
                    positions.secondElementAnchorPointYInCanvas + (ICON_SIZE / 2)
                ]

                line = [
                    positions.firstElementAnchorPointXInCanvas,
                    positions.firstElementAnchorPointYInCanvas]

                if (positions.selfReferenceAdditionalPoints != null) {
                    line = line.concat(positions.selfReferenceAdditionalPoints)
                }

                line = line.concat([
                    positions.secondElementAnchorPointXInCanvas,
                    positions.secondElementAnchorPointYInCanvas + ICON_SIZE
                ])
                break;
        }


        const firstMultiplicityPosition = {
            left: positions.firstElementAnchorPointXInCanvas,
            top: positions.firstElementAnchorPointYInCanvas,
        }

        const secondMultiplicityPosition = {
            left: positions.secondElementAnchorPointXInCanvas,
            top: positions.secondElementAnchorPointYInCanvas
        }

        const firstMultiplicityId = connection.id + "-firstMultiplicity"
        const secondMultiplicityId = connection.id + "-secondMultiplicity"

        const self = this
        const clickFunc = function (event: any) {
            self.onClickedOnObjectOnTheCanvas(event, connection, UmlConnectionType.RELATION)
        }

        var mouseDownOnFirstAnchorPointFunc = null
        var mouseDownOnSecondAnchorPointFunc = null
        if (selected) {
            mouseDownOnFirstAnchorPointFunc = event => self.onMouseDownOnAnchorPoint(event, connection.id, true, positions)
            mouseDownOnSecondAnchorPointFunc = event => self.onMouseDownOnAnchorPoint(event, connection.id, false, positions)
        }

        return (
            <Group key={connection.id}>
                <Line // workaround to increase the touchable area of a line
                    points={line}
                    strokeWidth={LINE_TOUCH_AREA_INCREASEMENT}
                    stroke={transparentLineColor}
                    onClick={clickFunc}
                />
                <Line
                    points={line}
                    strokeWidth={LINE_STROKE_WIDTH}
                    stroke={lineColor}
                    onClick={clickFunc}
                />
                <Line
                    points={icon}
                    closed={true}
                    fill={iconFill}
                    strokeWidth={LINE_STROKE_WIDTH}
                    stroke={lineColor}
                    onClick={clickFunc}
                />


                {selected &&
                    <Circle
                        x={positions.firstElementAnchorPointXInCanvas}
                        y={positions.firstElementAnchorPointYInCanvas}
                        radius={4}
                        stroke={"black"}
                        fill={"white"}
                        strokeWidth={4}
                        onMouseDown={mouseDownOnFirstAnchorPointFunc}
                    />
                }
                {selected &&
                    <Circle
                        x={positions.secondElementAnchorPointXInCanvas}
                        y={positions.secondElementAnchorPointYInCanvas}
                        radius={4}
                        stroke={"black"}
                        fill={"white"}
                        strokeWidth={4}
                        onMouseDown={mouseDownOnSecondAnchorPointFunc}

                    />
                }
            </Group>
        )

    }



    renderAssociation(connection: UmlConnection, positions: Positions, lineColor: string, selected: boolean): JSX.Element {

        var linePos: Array<number> = [
            positions.firstElementAnchorPointXInCanvas,
            positions.firstElementAnchorPointYInCanvas
        ]
        if (positions.selfReferenceAdditionalPoints != null) {
            linePos = linePos.concat(positions.selfReferenceAdditionalPoints)
        }


        linePos = linePos.concat([
            positions.secondElementAnchorPointXInCanvas,
            positions.secondElementAnchorPointYInCanvas
        ])

        const self = this
        const clickFunc = event => self.onClickedOnObjectOnTheCanvas(event, connection, UmlConnectionType.RELATION)
        var mouseDownOnFirstAnchorPointFunc = null
        var mouseDownOnSecondAnchorPointFunc = null
        if (selected) {
            mouseDownOnFirstAnchorPointFunc = event => self.onMouseDownOnAnchorPoint(event, connection.id, true, positions)
            mouseDownOnSecondAnchorPointFunc = event => self.onMouseDownOnAnchorPoint(event, connection.id, false, positions)
        }

        return (
            <Group key={connection.id}>
                <Line // workaround to increase the touchable area of a line
                    points={linePos}
                    strokeWidth={LINE_TOUCH_AREA_INCREASEMENT}
                    stroke={transparentLineColor}
                    onClick={clickFunc}
                />
                <Line
                    points={linePos}
                    strokeWidth={LINE_STROKE_WIDTH}
                    stroke={lineColor}
                    onClick={clickFunc}
                />


                {selected &&
                    <Circle
                        x={positions.firstElementAnchorPointXInCanvas}
                        y={positions.firstElementAnchorPointYInCanvas}
                        radius={4}
                        stroke={"black"}
                        fill={"white"}
                        strokeWidth={4}
                        onMouseDown={mouseDownOnFirstAnchorPointFunc}
                    />
                }
                {selected &&
                    <Circle
                        x={positions.secondElementAnchorPointXInCanvas}
                        y={positions.secondElementAnchorPointYInCanvas}
                        radius={4}
                        stroke={"black"}
                        fill={"white"}
                        strokeWidth={4}
                        onMouseDown={mouseDownOnSecondAnchorPointFunc}

                    />
                }
            </Group>
        )
    }



    areDependingElementsReady(): boolean {
        // TODO implement properly

        if (!this.props.umlDocument.types.isEmpty()){
            return this.getElement(this.props.umlDocument.types.first().id) != null
        }

        return true
    }



    runRenderingOnceDependingUmlElementsAreReady() {

        if (this.renderSubscription !== null) {
            this.renderSubscription.unsubscribe()
            this.renderSubscription = null
        }

        const ready = this.areDependingElementsReady()
        if (ready) {
            return this.renderCanvas()
        } else {
            // Elements not layouted yet
            // So schedule a redraw a few milliseconds later
            const self = this
            // TODO implement this properly
            this.renderSubscription = Rx.Observable.timer(50)
                .filter(ignored => self.areDependingElementsReady())
                .take(1)
                .subscribe(
                next => {
                    console.log("READY")
                },
                error => { },
                () => this.forceUpdate() // causes rerendering 
                )

            return null
        }
    }

    private getParentContainer(): HTMLElement {
        if (this.parentContainer == null) {
            this.parentContainer = document.getElementById("uml-canvas")
        }
        return this.parentContainer
    }

    // workaround. First we catch the click event on the canvas and create the corresponding intent.
    // but we dont fire the intent, because we have stop propagating the real click event in DOM.
    // See onClickedOnContainer
    onClickedOnObjectOnTheCanvas(event: any, connection: UmlConnection, connectionType: UmlConnectionType) {
        // console.log("CLICK_OBJ")
        // event.cancelBubble = true // konva canvas event handled. no other shapes on canvas receive this event
        switch (connectionType) {
            case UmlConnectionType.INHERITANCE:
                this.intentToDeliver = new SelectUmlInheritanceIntent(connection.id)
                break;
            case UmlConnectionType.INTERFACE_IMPLEMENTATION: {
                this.intentToDeliver = new SelectUmlInterfaceImplementationIntent(connection.id)
                break;
            }
            case UmlConnectionType.RELATION:
                this.intentToDeliver = new SelectUmlRelationIntent(connection.id)
                break;
            default: throw new Error("Unexpected UmlConnectionType: " + connectionType)
        }

    }

    onDoubleClickedOnObjectOnTheCanvas(event: any) {
        /*
        console.log("DOUBLE_CLICK_OBJ")
        event.cancleBubble = true // just copy pasted from above
    
        const x = event.pageX - event.target.offsetLeft
        const y = event.pageY - event.target.offsetTop
        this.intentToDeliver = new DoubleClickUmlConnectionIntent(this.props.connection.id, x, y, this.props.type)
        */
    }


    onClickedOnContainer(event: any) {
        if (this.intentToDeliver != null) {
            event.stopPropagation()
            this.props.intentEmitter.emit(this.intentToDeliver)
            this.intentToDeliver = null
        }

        if (this.mouseUpDeliveredBeforeCatchingClick) {
            event.stopPropagation()
            this.mouseUpDeliveredBeforeCatchingClick = false
        }
    }

    onDoubleClickedOnContainer(event: any) {
        /*
        console.log("DOUBLE_CLICK_CONTAINER")
        if (this.intentToDeliver != null) {
            event.stopPropagation()
            this.props.intentEmitter.emit(this.intentToDeliver)
            this.intentToDeliver = null
        }
        */
    }

    onMouseDownOnAnchorPoint(event: any, umlConnectionid: string, firstAnchorPoint: boolean, positions: Positions) {
        const parentContainer = this.getParentContainer()

        var max = 0
        var min = 0
        var achsis = true

        if (positions.firstAnchorPointSide === AnchorPointSide.LEFT || positions.firstAnchorPointSide === AnchorPointSide.RIGHT) {
            achsis = false
            min = firstAnchorPoint === true ? positions.firstElementY : positions.secondElementY
            max = firstAnchorPoint === true ? (positions.firstElementY + positions.firstElementHeight) : (positions.secondElementY + positions.secondElementHeight)
        } else {
            achsis = true
            min = firstAnchorPoint === true ? positions.firstElementX : positions.secondElementX
            max = firstAnchorPoint === true ? (positions.firstElementX + positions.firstElementWidth) : (positions.secondElementX + positions.secondElementWidth)
        }

        this.intentToDeliver = new StartMovingAnchorPointIntent(
            umlConnectionid,
            firstAnchorPoint,
            min,
            max,
            achsis
        )
        event.cancelBubble = true
    }

    onMouseMoveInCanvas(event: any) {
        const movingTransaction = this.props.movingAnchorPointInteractionTransaction
        if (movingTransaction != null) {
            const parentContainer = this.getParentContainer()
            var percentage = 0
            if (movingTransaction.achsis === true) {
                // X Achsis
                const mouseX = event.pageX - parentContainer.offsetLeft + parentContainer.scrollLeft
                percentage = mapValue(mouseX, movingTransaction.minMouseValue, movingTransaction.maxMouseValue, 0, 100)
            } else {
                // Y Achsis
                const mouseY = event.pageY - parentContainer.offsetTop + parentContainer.scrollTop
                percentage = mapValue(mouseY, movingTransaction.minMouseValue, movingTransaction.maxMouseValue, 0, 100)
            }

            this.props.intentEmitter.emit(MoveAnchorPointIntent.obtain(movingTransaction.umlConnectionId, percentage, movingTransaction.firstAnchorPoint))
            event.stopPropagation()
        }
    }

    onMouseDownInCanvas(event: any) {
        if (this.intentToDeliver != null) {
            this.props.intentEmitter.emit(this.intentToDeliver)
            this.intentToDeliver = null
            event.stopPropagation()
        }
    }

    onMouseUpInCanvas(event: any) {
        if (this.props.movingAnchorPointInteractionTransaction != null) {
            this.props.intentEmitter.emit(new StopMovingAnchorPointIntent())
            event.stopPropagation()
            this.mouseUpDeliveredBeforeCatchingClick = true
        } else {
            this.mouseUpDeliveredBeforeCatchingClick = false
        }
    }

    onTextInputRequiresFocus(id: string, idType: string) {
        const intentEmitter = this.props.intentEmitter;
        if (idType === FIRST_MULTIPLICITY_INPUT) {
            intentEmitter.emit(new StartEditingUmlRelationTextFieldIntent(id, UmlRelationTextFieldType.START_MULTIPLICITY))
        }
        else if (idType === SECOND_MULTIPLICITY_INPUT) {
            intentEmitter.emit(new StartEditingUmlRelationTextFieldIntent(id, UmlRelationTextFieldType.END_MULTIPLICITY))
        }
        else throw new Error("onTextInputRequiresFocus() on unknown idType " + idType)
    }

    onTextInputClicked(id: string, idType: string) {
        const intentEmitter = this.props.intentEmitter;
        this.props.intentEmitter.emit(new SelectUmlRelationIntent(id))
    }


    onTextInputChanged(id: string, idType: string, newText: string) {

        const intentEmitter = this.props.intentEmitter;
        if (idType === FIRST_MULTIPLICITY_INPUT) {
            intentEmitter.emit(new ChangeUmlRelationTextFieldIntent(id, UmlRelationTextFieldType.START_MULTIPLICITY, newText))
        }
        else if (idType === SECOND_MULTIPLICITY_INPUT) {
            intentEmitter.emit(new ChangeUmlRelationTextFieldIntent(id, UmlRelationTextFieldType.END_MULTIPLICITY, newText))
        }

        else throw new Error("onTextInputRequiresFocus() on unknown idType " + idType)

    }

    onTextInputCanceled(id: string, idType: string) {
        const intentEmitter = this.props.intentEmitter;
        if (idType === FIRST_MULTIPLICITY_INPUT) {
            intentEmitter.emit(new CancelEditingUmlRelationTextFieldIntent(id, UmlRelationTextFieldType.START_MULTIPLICITY))
        }
        else if (idType === SECOND_MULTIPLICITY_INPUT) {
            intentEmitter.emit(new CancelEditingUmlRelationTextFieldIntent(id, UmlRelationTextFieldType.START_MULTIPLICITY))
        }
        else throw new Error("onTextInputRequiresFocus() on unknown idType " + idType)
    }
}


/**
 * Ckaczkate the positoins required to draw the UML Relation
 * 
 * @param firstNode The first HTML node of the uml relation
 * @param secondNode The second HTML node of the uml relation
 * @param firstAnchorPoint The first Anchor Point (the posistion where rleation should be connection to firstNode)
 * @param secondAnchorPoint The second Anchor Point (the position where rleation should be connected to secondNode)
 */
function calculatePosition(firstNode: HTMLElement, secondNode: HTMLElement, firstAnchorPoint: AnchorPoint, secondAnchorPoint: AnchorPoint): Positions {

    const leftMost = firstNode.offsetLeft < secondNode.offsetLeft ? firstNode : secondNode
    const bottomMost = firstNode.offsetTop + firstNode.offsetHeight > secondNode.offsetTop + secondNode.offsetHeight ? firstNode : secondNode
    const rightMost = leftMost == firstNode ? secondNode : firstNode
    const topMost = firstNode.offsetTop < secondNode.offsetTop ? firstNode : secondNode

    const firstAnchorPointPercentage = firstAnchorPoint.percentage
    const secondAnchorPointPercentage = secondAnchorPoint.percentage


    if (firstNode == secondNode) {
        // It is a refernce to itself
        let firstAnchorPointSide = AnchorPointSide.TOP
        let secondAnchorPointSide = AnchorPointSide.RIGHT
        let [firstAnchorX, firstAnchorY] = calculateAnchorPointRelativeInCanvas(firstNode, firstAnchorPointSide, firstAnchorPointPercentage)
        let [secondAnchorX, secondAnchorY] = calculateAnchorPointRelativeInCanvas(firstNode, secondAnchorPointSide, secondAnchorPointPercentage)

        // Assumption: AnchorPoints side are TOP and RIGHT
        const additionalSelfRefPoints = [
            firstAnchorX,
            firstAnchorY - ICON_SIZE,
            firstNode.offsetLeft + firstNode.offsetWidth + ICON_SIZE,
            firstNode.offsetTop - ICON_SIZE,
            secondAnchorX + ICON_SIZE,
            secondAnchorY
        ]

        return {
            firstElementX: firstNode.offsetLeft,
            firstElementY: firstNode.offsetTop,
            firstElementWidth: firstNode.offsetWidth,
            firstElementHeight: firstNode.offsetHeight,
            secondElementX: secondNode.offsetLeft,
            secondElementY: secondNode.offsetTop,
            secondElementWidth: secondNode.offsetWidth,
            secondElementHeight: secondNode.offsetHeight,

            firstElementAnchorPointXInCanvas: firstAnchorX,
            firstElementAnchorPointYInCanvas: firstAnchorY,
            firstAnchorPointSide: firstAnchorPointSide,
            secondElementAnchorPointXInCanvas: secondAnchorX,
            secondElementAnchorPointYInCanvas: secondAnchorY,
            secondAnchorPointSide: secondAnchorPointSide,
            selfReferenceAdditionalPoints: additionalSelfRefPoints
        }
    }


    const [firstAnchorPointSide, secondAnchorPointSide] = anchorPointsWithMinimumDistance(firstNode, secondNode, firstAnchorPointPercentage, secondAnchorPointPercentage)



    const [firstAnchorX, firstAnchorY] = calculateAnchorPointRelativeInCanvas(firstNode, firstAnchorPointSide, firstAnchorPointPercentage)
    const [secondAnchorX, secondAnchorY] = calculateAnchorPointRelativeInCanvas(secondNode, secondAnchorPointSide, secondAnchorPointPercentage)


    return {
        firstElementX: firstNode.offsetLeft,
        firstElementY: firstNode.offsetTop,
        firstElementWidth: firstNode.offsetWidth,
        firstElementHeight: firstNode.offsetHeight,
        secondElementX: secondNode.offsetLeft,
        secondElementY: secondNode.offsetTop,
        secondElementWidth: secondNode.offsetWidth,
        secondElementHeight: secondNode.offsetHeight,

        firstElementAnchorPointXInCanvas: firstAnchorX,
        firstElementAnchorPointYInCanvas: firstAnchorY,
        firstAnchorPointSide: firstAnchorPointSide,
        secondElementAnchorPointXInCanvas: secondAnchorX,
        secondElementAnchorPointYInCanvas: secondAnchorY,
        secondAnchorPointSide: secondAnchorPointSide,
        selfReferenceAdditionalPoints: null
    }
}


/**
    Calculate the absolute position of the AnchorPoint
 * @param node The Htlm element the anchor point is attached to
 * @param anchorSide The AnchorPointSide
 * @param anchorPercentag The percentage of the anchor point
     */
function calculateAnchorPointRelativeInCanvas(node: HTMLElement, anchorSide: AnchorPointSide, anchorPercentag: number): [number, number] {
    const heightProportion = (node.offsetHeight / 100) * anchorPercentag
    const widthProportion = (node.offsetWidth / 100) * anchorPercentag

    var anchorX = 0
    var anchorY = 0

    switch (anchorSide) {
        case AnchorPointSide.LEFT:
            anchorX = node.offsetLeft
            anchorY = node.offsetTop + heightProportion
            break;
        case AnchorPointSide.RIGHT:
            anchorX = node.offsetLeft + node.offsetWidth
            anchorY = node.offsetTop + heightProportion
            break;
        case AnchorPointSide.TOP:
            anchorX = node.offsetLeft + widthProportion
            anchorY = node.offsetTop
            break;
        case AnchorPointSide.BOTTOM:
            anchorX = node.offsetLeft + widthProportion
            anchorY = (node.offsetTop + node.offsetHeight)
            break;
    }

    return [anchorX, anchorY]
}


/**
 * Calculates the distance between two coordinates (x1/y1) and (x2/y2)
 * @param x1 The x1 coordinate
 * @param y1 The y1 coordinate
 * @param x2 The x2 coordinate
 * @param y2 The y2 coordinate
 */
function distanceBetween(x1: number, y1: number, x2: number, y2: number): number {
    return Math.sqrt(Math.pow((x2 - x1), 2) + Math.pow(y2 - y1, 2))
}

/**
 * Calculates the sides of the two elements with the minimum distance between them
 */
function anchorPointsWithMinimumDistance(firstNode: HTMLElement, secondNode: HTMLElement, firstAnchorPointPercentage: number, secondAnchorPointPercentage: number): [AnchorPointSide, AnchorPointSide] {

    const [firstTopX, firstTopY] = calculateAnchorPointRelativeInCanvas(firstNode, AnchorPointSide.TOP, firstAnchorPointPercentage)
    const [firstRightX, firstRightY] = calculateAnchorPointRelativeInCanvas(firstNode, AnchorPointSide.RIGHT, firstAnchorPointPercentage)
    const [firstBottomX, firstBottomY] = calculateAnchorPointRelativeInCanvas(firstNode, AnchorPointSide.BOTTOM, firstAnchorPointPercentage)
    const [firstLeftX, firstLeftY] = calculateAnchorPointRelativeInCanvas(firstNode, AnchorPointSide.LEFT, firstAnchorPointPercentage)

    const [secondTopX, secondTopY] = calculateAnchorPointRelativeInCanvas(secondNode, AnchorPointSide.TOP, secondAnchorPointPercentage)
    const [secondRightX, secondRightY] = calculateAnchorPointRelativeInCanvas(secondNode, AnchorPointSide.RIGHT, secondAnchorPointPercentage)
    const [secondBottomX, secondBottomY] = calculateAnchorPointRelativeInCanvas(secondNode, AnchorPointSide.BOTTOM, secondAnchorPointPercentage)
    const [secondLeftX, secondLeftY] = calculateAnchorPointRelativeInCanvas(secondNode, AnchorPointSide.LEFT, secondAnchorPointPercentage)

    const firstTopSecondTop = distanceBetween(firstTopX, firstTopY, secondTopX, secondTopY)
    const firstTopSecondLeft = distanceBetween(firstTopX, firstTopY, secondLeftX, secondLeftY)
    const firstTopSecondBottom = distanceBetween(firstTopX, firstTopY, secondBottomX, secondBottomY)
    const firstTopSecondRight = distanceBetween(firstTopX, firstTopY, secondRightX, secondRightY)

    const firstLeftSecondTop = distanceBetween(firstLeftX, firstLeftY, secondTopX, secondTopY)
    const firstLeftSecondLeft = distanceBetween(firstLeftX, firstLeftY, secondLeftX, secondLeftY)
    const firstLeftSecondBottom = distanceBetween(firstLeftX, firstLeftY, secondBottomX, secondBottomY)
    const firstLeftSecondRight = distanceBetween(firstLeftX, firstLeftY, secondRightX, secondRightY)

    const firstBottomSecondTop = distanceBetween(firstBottomX, firstBottomY, secondTopX, secondTopY)
    const firstBottomSecondLeft = distanceBetween(firstBottomX, firstBottomY, secondLeftX, secondLeftY)
    const firstBottomSecondBottom = distanceBetween(firstBottomX, firstBottomY, secondBottomX, secondBottomY)
    const firstBottomSecondRight = distanceBetween(firstBottomX, firstBottomY, secondRightX, secondRightY)

    const firstRightSecondTop = distanceBetween(firstRightX, firstRightY, secondTopX, secondTopY)
    const firstRightSecondLeft = distanceBetween(firstRightX, firstRightY, secondLeftX, secondLeftY)
    const firstRightSecondBottom = distanceBetween(firstRightX, firstRightY, secondBottomX, secondBottomY)
    const firstRightSecondRight = distanceBetween(firstRightX, firstRightY, secondRightX, secondRightY)


    const distances: Array<number> = [
        firstTopSecondTop, firstTopSecondLeft, firstTopSecondBottom, firstTopSecondRight,
        firstLeftSecondTop, firstLeftSecondLeft, firstLeftSecondBottom, firstLeftSecondRight,
        firstBottomSecondTop, firstBottomSecondLeft, firstBottomSecondBottom, firstBottomSecondRight,
        firstRightSecondTop, firstRightSecondLeft, firstRightSecondBottom, firstRightSecondRight
    ]

    var minValue = Number.MAX_VALUE;
    var minIndex = 0;

    distances.forEach((d, index) => {
        if (d < minValue) {
            minValue = d
            minIndex = index
        }
    });

    return ANCHOR_POINT_LOOKUP[minIndex]
}


function mapValue(x: number, sourceStart: number, sourceEnd: number, targetStart: number, targetEnd: number): number {

    if (x <= sourceStart) {
        return targetStart;
    }

    if (x >= sourceEnd) {
        return targetEnd;
    }

    return (x - sourceStart) / (sourceEnd - sourceStart)
        * (targetEnd - targetStart) + targetStart;
}